from unittest.mock import MagicMock
import geopandas as gpd
import pandas as pd
from shapely.geometry import Polygon

import pytest
from autocnet.control import control

def test_identify_potential_overlaps(controlnetwork, candidategraph):
    res = control.identify_potential_overlaps(candidategraph,
                                              controlnetwork,
                                              overlap=False)

    assert res.equals(pd.Series([(2,), (2,),
                                 (1,), (1,),
                                 (0,), (0,)],
                                 index=[6,7,8,9,10,11]))

def test_potential_overlap(controlnetwork, candidategraph):
    # Patch in an intersection check so that all points intersect all geoms
    candidategraph.create_node_subgraph = MagicMock(return_value=candidategraph)
    coords = [(-1., -1.), (-1., 1.), (1., 1.), (1., -1.), (-1., -1.)]
    poly = gpd.GeoSeries(Polygon(coords))
    candidategraph.compute_intersection = MagicMock(return_value=(poly, 0))
    res = control.identify_potential_overlaps(candidategraph,
                                              controlnetwork,
                                              overlap=True)

    assert res.equals(pd.Series([(2,), (2,),
                                 (1,), (1,),
                                 (0,), (0,)],
                                 index=[6,7,8,9,10,11]))

def test_compute_covariance():
    df = pd.DataFrame([[0,0,3], [0,0,4], [0,0,2]], columns=['adjustedY', 'adjustedX', 'pointtype'])
    df = control.compute_covariance(df, 10, 10, 15, 100)
    
    def assertexists(row):
        if row['pointtype'] > 2:
            assert row['aprioriCovar'].sum() > 100
        else:
            assert row['aprioriCovar'] == []

    df.apply(assertexists, axis=1)


"""
def test_fromcandidategraph(candidategraph, controlnetwork_data):#, controlnetwork):
    matches = candidategraph.get_matches()
    cn = control.ControlNetwork.from_candidategraph(matches)
    assert cn.data[['point_id', 'image_index']].equals(controlnetwork_data[['point_id', 'image_index']])

def test_add_measure():
    cn = control.ControlNetwork()

    # Add the point 0 from image 0
    key = (0,0)
    cn.add_measure(key, (0,1), 3, [1,1])
    assert key in cn.measure_to_point.keys()
    assert cn.measure_to_point[key] == 0

    # Add the point 1 from image 2
    key = (1,2)
    cn.add_measure(key, (0,1), 2, [1,1])
    assert key in cn.measure_to_point.keys()

    # Add another measure associated with point (0,0)
    #  Key is the source and this methods is called to add the destination
    key = (2,1)
    cn.add_measure(key, (0,2), 3, [1,1], point_id = 0)
    assert key in cn.measure_to_point.keys()
    assert cn.measure_to_point[key] == 0

def test_validate_points(controlnetwork):
    assert not controlnetwork.validate_points().any()

def test_bad_validate_points(bad_controlnetwork):
    assert bad_controlnetwork.validate_points().iloc[0] == True
    assert not bad_controlnetwork.validate_points().iloc[1:].all()




"""


import os
import warnings
import yaml
import socket

from sqlalchemy import create_engine, pool, orm
from sqlalchemy.event import listen

from pkg_resources import get_distribution, DistributionNotFound

try:
    _dist = get_distribution('autocnet')
    # Normalize case for Windows systems
    dist_loc = os.path.normcase(_dist.location)
    here = os.path.normcase(__file__)
    if not here.startswith(os.path.join(dist_loc, 'autocnet')):
        # not installed, but there is another version that *is*
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = 'Please install this project with setup.py'
else:
    __version__ = _dist.version

# Patch the candidate graph into the root namespace
from autocnet.graph.network import CandidateGraph, NetworkCandidateGraph

import autocnet.examples
import autocnet.camera
import autocnet.cg
import autocnet.control
import autocnet.graph
import autocnet.matcher
import autocnet.transformation
import autocnet.utils
import autocnet.spatial

def get_data(filename):
    packagdir = autocnet.__path__[0]
    dirname = os.path.join(os.path.dirname(packagdir), 'data')
    fullname = os.path.join(dirname, filename)
    return fullname

def cuda(enable=False, gpu=0):
    # Classes/Methods that can vary if GPU is available
    from autocnet.graph.node import Node
    from autocnet.graph.edge import Edge
    if enable:
        try:
            import cudasift as cs
            cs.PyInitCuda(gpu)

            # Here is where the GPU methods get patched into the class
            from autocnet.matcher.cuda_extractor import extract_features
            Node._extract_features = staticmethod(extract_features)

            from autocnet.matcher.cuda_matcher import match
            Edge._match = staticmethod(match)

            from autocnet.matcher.cuda_decompose import decompose_and_match
            Edge.decompose_and_match = decompose_and_match

            from autocnet.matcher.cuda_outlier_detector import distance_ratio
            Edge._ratio_check = staticmethod(distance_ratio)

        except Exception:
            warnings.warn('Failed to enable Cuda')
        return

    # Here is where the CPU methods get patched into the class
    from autocnet.matcher.cpu_extractor import extract_features
    Node._extract_features = staticmethod(extract_features)

    from autocnet.matcher.cpu_matcher import match
    Edge._match = staticmethod(match)

    from autocnet.matcher.cpu_decompose import decompose_and_match
    Edge.decompose_and_match = decompose_and_match

    from autocnet.matcher.cpu_outlier_detector import  distance_ratio
    Edge._ratio_check = staticmethod(distance_ratio)

cuda()



from datetime import datetime
import json
import sys

import numpy as np
import pandas as pd
import pytest
import sqlalchemy
from unittest.mock import MagicMock, patch

from autocnet.io.db import model
from autocnet.graph.network import NetworkCandidateGraph

from shapely.geometry import MultiPolygon, Polygon, Point

if sys.platform.startswith("darwin"):
    pytest.skip("skipping DB tests for MacOS", allow_module_level=True)

def test_keypoints_exists(tables):
    assert model.Keypoints.__tablename__ in tables

def test_edges_exists(tables):
    assert model.Edges.__tablename__ in tables

def test_costs_exists(tables):
    assert model.Costs.__tablename__ in tables

def test_matches_exists(tables):
    assert model.Matches.__tablename__ in tables

def test_cameras_exists(tables):
    assert model.Cameras.__tablename__ in tables

def test_measures_exists(tables):
    assert model.Measures.__tablename__ in tables

def test_points_history_exists(tables):
    assert model.PointsHistory.__tablename__ in tables

def test_measures_history_exists(tables):
    assert model.MeasuresHistory.__tablename__ in tables

def test_job_history_exists(tables):
    assert model.JobsHistory.__tablename__ in tables


def test_create_camera_without_image(session):
    with pytest.raises(sqlalchemy.exc.IntegrityError):
        model.Cameras.create(session, **{'image_id':1})

def test_create_camera(session):
    #with pytest.raises(sqlalchemy.exc.IntegrityError):
    c = model.Cameras.create(session)
    res = session.query(model.Cameras).first()
    assert c.id == res.id

def test_create_camera_unique_constraint(session):
    model.Images.create(session, **{'id':1})
    data = {'image_id':1}
    model.Cameras.create(session, **data)
    with pytest.raises(sqlalchemy.exc.IntegrityError):
        model.Cameras.create(session, **data)

def test_images_exists(tables):
    assert model.Images.__tablename__ in tables

@pytest.mark.parametrize('data', [
    {'id':1},
    {'name':'foo',
     'path':'/neither/here/nor/there'},
    ])
def test_create_images(session, data):
    i = model.Images.create(session, **data)
    resp = session.query(model.Images).filter(model.Images.id==i.id).first()
    assert i == resp

'''@pytest.mark.parametrize('data', [
    {'id':1},
    {'serial':'foo'}
])
def test_create_images_constrined(session, data):
    """
    Test that the images unique constraint is being observed.
    """
    model.Images.create(session, **data)
    with pytest.raises(sqlalchemy.exc.IntegrityError):
        model.Images.create(session, **data)'''

def test_overlay_exists(tables):
    assert model.Overlay.__tablename__ in tables

@pytest.mark.parametrize('data', [
    {'id':1},
    {'id':1, 'intersections':[1,2,3]},
    {'id':1, 'intersections':[1,2,3],
     'geom':Polygon([(0,0), (1,0), (1,1), (0,1), (0,0)])}

])
def test_create_overlay(session, data):
    d = model.Overlay.create(session, **data)
    resp = session.query(model.Overlay).filter(model.Overlay.id == d.id).first()
    assert d == resp

def test_points_exists(tables):
    assert model.Points.__tablename__ in tables

@pytest.mark.parametrize("data", [
    {'id':1, 'pointtype':2},
    {'pointtype':2, 'identifier':'123abc'},
    {'pointtype':3, 'apriori':Point(0,0,0)},
    {'pointtype':3, 'adjusted':Point(0,0,0)},
    {'pointtype':2, 'adjusted':Point(1,1,1), 'ignore':False}
])
def test_create_point(session, data):
    p = model.Points.create(session, **data)
    resp = session.query(model.Points).filter(model.Points.id == p.id).first()
    assert p == resp

@pytest.mark.parametrize("data, expected", [
    ({'pointtype':3, 'adjusted':Point(0,-1000000,0)}, Point(270, 0)),
    ({'pointtype':3}, None)
])
def test_create_point_geom(session, data, expected):
    p = model.Points.create(session, **data)
    resp = session.query(model.Points).filter(model.Points.id == p.id).first()

    assert resp.geom == expected

@pytest.mark.parametrize("data, new_adjusted, expected", [
    ({'pointtype':3, 'adjusted':Point(0,-100000,0)}, None, None),
    ({'pointtype':3, 'adjusted':Point(0,-100000,0)}, Point(0,100000,0), Point(90, 0)),
    ({'pointtype':3}, Point(0,-100000,0), Point(270, 0))
])
def test_update_point_geom(session, data, new_adjusted, expected):
    p = model.Points.create(session, **data)
    p.adjusted = new_adjusted
    session.commit()
    resp = session.query(model.Points).filter(model.Points.id == p.id).first()
    assert resp.geom == expected

# def test_point_trigger(session):
#     original = 3
#     new_type = 2

#     data = {'pointtype':original, 'adjusted' : Point(1,10000,1)}

#     with session as s:
#         p = model.Points.create(s, **data)

#         p.pointtype = new_type
#         # p.pointtype(new_type)
#         s.commit()
#         s.delete(p)
#         s.commit()

#         resp = session.query(model.PointsHistory).filter(model.PointsHistory.fk == p.id).first()
#         print(f"\nstart of debugging output for test_point_trigger\n_________________________________________\nthe value of p is: {p}\n----\nthe value of p.id is: {p.id}\n----\nthe value of resp is below\n----\n{resp}\n_________________________________________\nEnd of debugging output for test_point_trigger\n")

#         assert resp.event == "insert"
#         assert resp[0].before == None
#         assert resp[0].after["pointType"] == original 

#         assert resp[1].event == "update"
#         assert resp[1].before['pointType'] == original
#         assert resp[1].after["pointType"] == new_type

#         assert resp[2].event == "delete"
#         assert resp[2].before['pointType'] == new_type
#         assert resp[2].after == None
#         s.close()

@pytest.mark.xfail(reason="Unknown issue on GitHub actions, passes locally")
def test_measure_trigger(session):
    original = 3
    new_type = 2

    with session as s:     
        point_data = {'pointtype':original, 'adjusted' : Point(1,10000,1)}
        p = model.Points.create(s, **point_data)

        measure_data = {'id' : 100 ,'sample': original, 'line': 10, 'pointid': p.id, 'serial': 'measure', '_measuretype' : 2}
        m = model.Measures(**measure_data)
        s.add(m)
        s.commit()

        measures = s.query(model.Measures).all()
        points = s.query(model.Points).all()
        print(measures, len(measures))
        print(points, len(points))

        m.sample = new_type
        s.commit()
        s.delete(p) 
        s.delete(m)
        s.commit()
 
    measures = session.query(model.Measures).all()
    points = session.query(model.Points).all()
    print(measures, len(measures))
    print(points, len(points))

    resp = session.query(model.MeasuresHistory).filter(model.MeasuresHistory.fk == m.id).all()
    print(resp, len(resp))
    assert resp[0].event == "insert"
    assert resp[0].before == None
    assert resp[0].after["sample"] == original 

    assert resp[1].event == "update"
    assert resp[1].before['sample'] == original
    assert resp[1].after["sample"] == new_type

    assert resp[2].event == "delete"
    assert resp[2].before['sample'] == new_type
    assert resp[2].after == None

def test_null_footprint(session):
    i = model.Images.create(session, geom=None,
                                      serial = 'serial')
    assert i.geom is None

def test_broken_bad_geom(session):
    # An irreperablly damaged poly
    truthgeom = MultiPolygon([Polygon([(0,0), (1,1), (1,2), (1,1), (0,0)])])
    i = model.Images.create(session, geom=truthgeom,
                                      serial = 'serial')
    resp = session.query(model.Images).filter(model.Images.id==i.id).one()
    assert resp.ignore == True

def test_fix_bad_geom(session):
    truthgeom = MultiPolygon([Polygon([(0,0), (0,1), (1,1), (0,1), (1,1), (1,0), (0,0) ])])
    i = model.Images.create(session, geom=truthgeom,
                                     serial = 'serial')
    resp = session.query(model.Images).filter(model.Images.id==i.id).one()
    assert resp.ignore == False
    assert resp.geom == MultiPolygon([Polygon([(0,1), (1,1), (1,0), (0,0), (0,1)])])

@pytest.mark.parametrize("measure_data, point_data, image_data", [(
    [{'id': 1, 'pointid': 1, 'imageid': 1, 'serial': 'ISISSERIAL1', 'measuretype': 3, 'sample': 0, 'line': 0},
     {'id': 2, 'pointid': 1, 'imageid': 2, 'serial': 'ISISSERIAL2', 'measuretype': 3, 'sample': 0, 'line': 0},
     {'id': 3, 'pointid': 1, 'imageid': 3, 'serial': 'ISISSERIAL3', 'measuretype': 3, 'sample': 0, 'line': 0},
     {'id': 4, 'pointid': 1, 'imageid': 4, 'serial': 'ISISSERIAL4', 'measuretype': 3, 'sample': 0, 'line': 0}],
    {'id':1,
     'pointtype':2},
    [{'id':1, 'serial': 'ISISSERIAL1'},
     {'id':2, 'serial': 'ISISSERIAL2'},
     {'id':3, 'serial': 'ISISSERIAL3'},
     {'id':4, 'serial': 'ISISSERIAL4'}])])
def test_ignore_image(session, measure_data, point_data, image_data):
    for data in image_data:
        model.Images.create(session, **data)
    model.Points.create(session, **point_data)
    for data in measure_data:
        model.Measures.create(session, **data)
    image_resp = session.query(model.Images).filter(model.Images.id == 1).first()
    image_resp.ignore = True
    ignored_measures_resp = session.query(model.Measures).filter(model.Measures.ignore == True).first()
    assert ignored_measures_resp.imageid == 1
    valid_measures_resp = session.query(model.Measures).filter(model.Measures.ignore == False)
    assert valid_measures_resp.count() == 3



from plio.io.io_gdal import GeoDataset
import pytest

from autocnet.transformation import affine

def test_estimate_affine_transformation():
    gd_base = GeoDataset('tests/test_subpixel_match/B08_012650_1780_XN_02S046W.l1.cal.destriped.crop.cub')
    gd_match = GeoDataset('tests/test_subpixel_match/J04_046447_1777_XI_02S046W.l1.cal.destriped.crop.cub')
    affine_transform = affine.estimate_affine_from_sensors(gd_base, gd_match, 150, 150)
    assert affine_transform.rotation == pytest.approx(-0.0012609633370663982, 6)
    assert affine_transform.shear == pytest.approx(0.01262411827876344)
    assert affine_transform.scale[0] == pytest.approx(1.0, 6)
    assert affine_transform.scale[1] == pytest.approx(1.0000000000000002, 6) 
    assert affine_transform.translation[0] == pytest.approx(10.591885552140951, 6)
    assert affine_transform.translation[1] == pytest.approx(1.7377773467863733, 6)


from pathlib import PosixPath
from psycopg2.extensions import register_adapter, AsIs, adapt

# Adaptors as defined by https://www.psycopg.org/docs/advanced.html#adapting-new-python-types-to-sql-syntax

def adapt_posixpath(posixpath):
    stringified = adapt(str(posixpath))
    return stringified

# TODO: Add adaptors from numpy types. We shouldn't have to typecast 
#  explicitly across the code base.

register_adapter(PosixPath, adapt_posixpath)


import logging

import numpy as np
import pandas as pd

import cv2

FLANN_INDEX_KDTREE = 1  # Algorithm to set centers,
DEFAULT_FLANN_PARAMETERS = dict(algorithm=FLANN_INDEX_KDTREE, trees=3)

log = logging.getLogger(__name__)

def match(edge, k=2, **kwargs):
    """
    Given two sets of descriptors, utilize a FLANN (Approximate Nearest
    Neighbor KDTree) matcher to find the k nearest matches.  Nearness is
    the euclidean distance between descriptors.

    The matches are then added as an attribute to the edge object.

    Parameters
    ----------
    k : int
	The number of neighbors to find
    """

    def _add_matches(matches):
        """
        Given a dataframe of matches, either append to an existing
        matches edge attribute or initially populate said attribute.

        Parameters
        ----------
        matches : dataframe
                  A dataframe of matches
        """
        if edge.matches.empty:
            edge.matches = matches
        else:
            df = edge.matches
            edge.matches = df.append(matches,
                                     ignore_index=True,
                                     verify_integrity=True)

    def mono_matches(a, b, aidx=None, bidx=None):
        """
	    Apply the FLANN match_features

    	Parameters
    	----------
    	a : object
    	    A node object

    	b : object
    	    A node object

    	aidx : iterable
    		An index for the descriptors to subset

    	bidx : iterable
    		An index for the descriptors to subset
    	"""
    	# Subset if requested
        if aidx is not None:
            ad = a.descriptors[aidx]
        else:
            ad = a.descriptors

        if bidx is not None:
            bd = b.descriptors[bidx]
        else:
            bd = b.descriptors

        # Load, train, and match
        fl.add(ad, a['node_id'], index=aidx)
        fl.train()
        matches = fl.query(bd, b['node_id'], k, index=bidx)
        _add_matches(matches)
        fl.clear()

    fl = FlannMatcher()

    # Get the correct descriptors
    aidx = kwargs.pop('aidx', None)
    bidx = kwargs.pop('bidx', None)

    mono_matches(edge.source, edge.destination, aidx=aidx, bidx=bidx)
    # Swap the indices since mono_matches is generic and source/destin are
    # swapped
    mono_matches(edge.destination, edge.source, aidx=bidx, bidx=aidx)

    source_keypoints = edge.source.keypoints[['x', 'y']]
    source_keypoints.rename(columns={'x': 'source_x', 'y': 'source_y'}, inplace=True)
    edge.matches = edge.matches.join(source_keypoints, 'source_idx')

    destination_keypoints = edge.destination.keypoints[['x', 'y']]
    destination_keypoints.rename(columns={'x': 'destination_x', 'y': 'destination_y'}, inplace=True)
    edge.matches = edge.matches.join(destination_keypoints, 'destination_idx')
    edge.matches.sort_values(by=['distance'])



class FlannMatcher(object):
    """
    A wrapper to the OpenCV Flann based matcher class that adds
    metadata tracking attributes and methods.  This takes arbitrary
    descriptors and so should be available for use with any
    descriptor data stored as an ndarray.

    Attributes
    ----------
    image_indices : dict
                    with key equal to the train image index (returned by the DMatch object),
                    e.g. an integer array index
                    and value equal to the image identifier, e.g. the name

    image_index_counter : int
                          The current number of images loaded into the matcher
    """

    def __init__(self, flann_parameters=DEFAULT_FLANN_PARAMETERS):
        self._flann_matcher = cv2.FlannBasedMatcher(flann_parameters, {})
        self.nid_lookup = {}
        self.search_idx = {}
        self.node_counter = 0

    def add(self, descriptor, nid, index=None):
        """
        Add a set of descriptors to the matcher and add the image
        index key to the image_indices attribute

        Parameters
        ----------
        descriptor : ndarray
                     The descriptor to be added

        nid : int
              The node ids
        """
        self._flann_matcher.add([descriptor])
        self.nid_lookup[self.node_counter] = nid
        self.node_counter += 1
        if index is not None:
            self.search_idx = dict((i, j) for i, j in enumerate(index))
        else:
            self.search_idx = dict((i,i) for i in range(len(descriptor)))

    def clear(self):
        """
        Remove all nodes from the tree and resets
        all counters
        """
        self._flann_matcher.clear()
        self.nid_lookup = {}
        self.node_counter = 0
        self.search_idx = {}

    def train(self):
        """
        Using the descriptors, generate the KDTree
        """
        self._flann_matcher.train()

    def query(self, descriptor, query_image, k=3, index=None):
        """

        Parameters
        ----------
        descriptor : ndarray
                     The query descriptor to search for

        query_image : hashable
                      Key of the query image

        k : int
            The number of nearest neighbors to search for

        index : iterable
                An iterable of observation indices to utilize for
                the input descriptors

        Returns
        -------
        matched : dataframe
                  containing matched points with columns containing:
                  matched image name, query index, train index, and
                  descriptor distance
        """

        matches = self._flann_matcher.knnMatch(descriptor, k=k)
        matched = []
        for i, m in enumerate(matches):
            for j in m:
                if index is not None:
                    qid = index[i]
                else:
                    qid = j.queryIdx
                source = query_image
                destination = self.nid_lookup[j.imgIdx]
                if source < destination:
                    matched.append((query_image,
                                    qid,
                                    destination,
                                    self.search_idx[j.trainIdx],
                                    j.distance))
                elif source > destination:
                    matched.append((destination,
                                    self.search_idx[j.trainIdx],
                                    query_image,
                                    qid,
                                    j.distance))
                else:
                    log.warning('Likely self neighbor in query!')
        return pd.DataFrame(matched, columns=['source_image', 'source_idx',
                                              'destination_image', 'destination_idx',
                                              'distance']).astype(np.float32)



from collections import defaultdict, MutableMapping, Counter
from functools import wraps, singledispatch
import json
import logging

import numpy as np
import pandas as pd
import networkx as nx
from scipy.spatial.distance import cdist
from shapely.geometry import Point
import sqlalchemy

from autocnet.graph.node import Node
from autocnet.utils import utils
from autocnet.matcher import cpu_outlier_detector as od
from autocnet.matcher import suppression_funcs as spf
from autocnet.matcher import subpixel as sp
from autocnet.matcher import cpu_ring_matcher
from autocnet.transformation import fundamental_matrix as fm
from autocnet.transformation import homography as hm
from autocnet.transformation.spatial import reproject, og2oc
from autocnet.vis.graph_view import plot_edge, plot_node, plot_edge_decomposition, plot_matches
from autocnet.cg import cg
from autocnet.io.db.model import Images, Keypoints, Matches,\
                                 Cameras, Base, Overlay, Edges,\
                                 Costs, Measures, Points, Measures
from autocnet.io.db.wrappers import DbDataFrame

from plio.io.io_gdal import GeoDataset

# set up the logging file
log = logging.getLogger(__name__)

class Edge(dict, MutableMapping):
    """
    Attributes
    ----------
    source : hashable
             The source node

    destination : hashable
                  The destination node
    masks : set
            A list of the available masking arrays

    weights : dict
             Dictionary with two keys overlap_area, and overlap_percn
             overlap_area returns the area overlaped by both images
             overlap_percn retuns the total percentage of overlap
    """

    def __init__(self, source=None, destination=None):
        self.source = source
        self.destination = destination
        self['homography'] = None
        self['fundamental_matrix'] = None
        self.subpixel_matches = pd.DataFrame()
        self._matches = pd.DataFrame()

        self['source_mbr'] = None
        self['destin_mbr'] = None
        self['overlap_latlon_coords'] = None

    def __repr__(self):
        return """
        Source Image Index: {}
        Destination Image Index: {}
        Available Masks: {}
        """.format(self.source, self.destination, self.masks)

    def __eq__(self, other):
        return utils.compare_dicts(self.__dict__, other.__dict__) *\
               utils.compare_dicts(self, other)

    @property
    def weights(self):
        if not hasattr(self, '_weights'):
            self._weights = {}
        return self._weights

    @weights.setter
    def weights(self, kv):
        key, value = kv
        self.weights[key] = value

    @property
    def masks(self):
        if not hasattr(self, '_masks'):
            self._masks = pd.DataFrame()
        return self._masks

    @masks.setter
    def masks(self, value):
        if isinstance(value, pd.DataFrame):
            self._masks = value
        else:
            raise(TypeError)

    @property
    def matches(self):
        if not hasattr(self, '_matches'):
            self._matches = pd.DataFrame()
        return self._matches

    @matches.setter
    def matches(self, value):
        if isinstance(value, pd.DataFrame):
            self._matches = value
            # Ensure that the costs df remains in sync with the matches df
            if not self.costs.index.equals(value.index):
                self.costs = pd.DataFrame(index=value.index)
        else:
            raise(TypeError)

    @property
    def costs(self):
        if not hasattr(self, '_costs'):
            self._costs = pd.DataFrame(index=self.matches.index)
        return self._costs

    @costs.setter
    def costs(self, value):
        if isinstance(value, pd.DataFrame):
            self._costs = value
        else:
            raise(TypeError)

    @property
    def ring(self):
        if not hasattr(self, '_ring'):
            self._ring = None
        return self._ring

    @ring.setter
    def ring(self, val):
        self._ring = val

    def match(self, k=2, **kwargs):

        """
        Given two sets of descriptors, utilize a FLANN (Approximate Nearest
        Neighbor KDTree) matcher to find the k nearest matches.  Nearness is
        the euclidean distance between descriptors.

        The matches are then added as an attribute to the edge object.

        Parameters
        ----------
        k : int
            The number of neighbors to find
        """
        # Reset the edge masks because matching is happening (again)
        self.masks = pd.DataFrame()
        kwargs['aidx'] = self.get_keypoints('source', overlap=True).index
        kwargs['bidx'] = self.get_keypoints('destination', overlap=True).index
        Edge._match(self, k=k, **kwargs)

    @staticmethod
    def _match(edge, k=2, **kwargs):
        """
        Patches the static cpu_matcher.match(edge) or cuda_match.match(edge)
        into the member method Edge.match()

        Parameters
        ----------
        edge : Edge
               The edge object to compute matches for; Edge.match() calls this
               with self
        k : int
            The number of neighbors to find
        """
        pass

    def ring_match(self, *args, **kwargs):
        ref_kps =  self.source.keypoints
        ref_desc = self.source.descriptors
        tar_kps = self.destination.keypoints
        tar_desc = self.destination.descriptors

        if not 'xm' in ref_kps.columns:
            log.warning('To ring match body centered coordinates (xm, ym, zm) must be in the keypoints')
            return
        ref_feats = ref_kps[['x', 'y', 'xm', 'ym', 'zm']].values
        tar_feats = tar_kps[['x', 'y', 'xm', 'ym', 'zm']].values

        _, _, pidx, ring = cpu_ring_matcher.ring_match(ref_feats, tar_feats,
                                                           ref_desc, tar_desc,
                                                           *args, **kwargs)

        if pidx is None:
            return
        self.ring = ring
        pidx = cpu_ring_matcher.check_pidx_duplicates(pidx)

        #Set the columns of the matches df
        matches = np.empty((pidx.shape[0], 4))
        matches[:,0] = self.source['node_id']
        matches[:,1] = ref_kps.index[pidx[:,0]].values
        matches[:,2] = self.destination['node_id']
        matches[:,3] = tar_kps.index[pidx[:,1]].values

        matches = pd.DataFrame(matches, columns=['source',
                                                 'source_idx',
                                                 'destination',
                                                 'destination_idx']).astype(np.float32)

        matches = matches.drop_duplicates()

        self.matches = matches

    def add_coordinates_to_matches(self):
        """
        Add source and destination x/y columns to the matches dataframe. This
        will add to the overall memory needed to store matches, but makes
        access to x,y easier as a join on the keypoints is not requires.
        """
        skps = self.get_keypoints(self.source, index=self.matches.source_idx)
        skps.reindex(self.matches['source_idx'])
        self.matches['source_x'] = skps.values[:,0]
        self.matches['source_y'] = skps.values[:,1]
        dkps = self.get_keypoints(self.destination, index=self.matches.destination_idx)
        dkps.reindex(self.matches['destination_idx'])
        self.matches['destination_x'] = dkps.values[:,0]
        self.matches['destination_y'] = dkps.values[:,1]

    def project_matches(self, semimajor, semiminor, on='source', srid=None):
        """
        Project matches.
        """
        try:
            coords = self.matches[['{}_y'.format(on),'{}_x'.format(on)]].values
        except:
            self.add_coordinates_to_matches()
            coords = self.matches[['{}_y'.format(on),'{}_x'.format(on)]].values

        node = getattr(self, on)
        camera = getattr(node, 'camera')
        if camera is None:
            log.warning('Unable to project matches without a sensor model.')
            return

        matches = self.matches

        gnd = np.empty((len(coords), 3))
        # Project the points to the surface and reproject into latlon space
        for i in range(gnd.shape[0]):
            ic = csmapi.ImageCoord(coords[i][0], coords[i][1])
            ground = camera.imageToGround(ic, 0)
            gnd[i] = [ground.x, ground.y, ground.z]
        lon_og, lat_og, alt = reproject(gnd.T, semimajor, semiminor,
                                    'geocent', 'latlon')
        lon, lat = og2oc(lon_og, lat_og, semimajor, semiminor)

        if srid:
            geoms = []
            for coord in zip(lon, lat, alt):
                geoms.append('SRID={};POINTZ({} {} {})'.format(srid, coord[0],
                                                                     coord[1],
                                                                     coord[2]))
            matches['geom'] = geoms

        matches['lat'] = lat
        matches['lon'] = lon
        self.matches = matches

    def decompose(self):
        """
        Apply coupled decomposition to the images and
        match identified sub-images
        """
        pass

    def decompose_and_match(*args, **kwargs):
        pass

    def overlap_check(self):
        """
        Creates a mask for matches on the overlap
        """
        if not (self["source_mbr"] and self["destin_mbr"]):
            log.warning(
                "Cannot use overlap constraint, minimum bounding rectangles"
                " have not been computed for one or more Nodes")
            return
        # Get overlapping keypts
        s_idx = self.get_keypoints(self.source, overlap=True).index
        d_idx = self.get_keypoints(self.destination, overlap=True).index
        # Create a mask from matches whose rows have both source idx &
        # dest idx in the overlapping keypts
        mask = pd.Series(False, index=self.matches.index)
        mask.loc[(self.matches["source_idx"].isin(s_idx)) &
                 (self.matches["destination_idx"].isin(d_idx))] = True
        self.masks['overlap'] = mask

    def symmetry_check(self, clean_keys=[], maskname='symmetry', **kwargs):
        self.masks[maskname] = od.mirroring_test(self.matches)

    def ratio_check(self, clean_keys=[], maskname='ratio', **kwargs):
        matches, mask = self.clean(clean_keys)
        self.masks[maskname] = self._ratio_check(self, matches, **kwargs)

    @staticmethod
    def _ratio_check(edge, matches, **kwargs):
        pass
        #return.masks[maskname] = od.distance_ratio(matches, **kwargs)

    @utils.methodispatch
    def get_keypoints(self, node, index=None, homogeneous=False, overlap=False):
        if not hasattr(index, '__iter__') and index is not None:
            raise TypeError
        keypts = node.get_keypoint_coordinates(index=index, homogeneous=homogeneous)
        # If the index is passed, the results are returned sorted. The index is not
        # necessarily sorted, so 'unsort' so that the return order matches the passed
        # order
        if index is not None:
            keypts = keypts.reindex(index)
        # If we only want keypoints in the overlap
        if overlap:
            if self.source == node:
                mbr = self['source_mbr']
            else:
                mbr = self['destin_mbr']
            # Can't use overlap if we haven't computed MBRs
            if mbr is None:
                return keypts
            return keypts.query('x >= {} and x <= {} and y >= {} and y <= {}'.format(*mbr))
        return keypts

    @get_keypoints.register(str)
    def _(self, node, index=None, homogeneous=False, overlap=False):
        if not hasattr(index, '__iter__') and index is not None:
            raise TypeError
        node = node.lower()
        node = getattr(self, node)
        return self.get_keypoints(node, index=index, homogeneous=homogeneous, overlap=overlap)

    def compute_fundamental_matrix(self, clean_keys=[], maskname='fundamental', **kwargs):
        """
        Estimate the fundamental matrix (F) using the correspondences tagged to this
        edge.


        Parameters
        ----------
        clean_keys : list
                     Of strings used to apply masks to omit correspondences

        method : {linear, nonlinear}
                 Method to use to compute F.  Linear is significantly faster at
                 the cost of reduced accuracy.

        See Also
        --------
        autocnet.transformation.transformations.FundamentalMatrix

        """
        _, mask = self.clean(clean_keys)
        s_keypoints, d_keypoints = self.get_match_coordinates(clean_keys=clean_keys)
        self.fundamental_matrix, fmask = fm.compute_fundamental_matrix(s_keypoints, d_keypoints, **kwargs)
        fmask = fmask.flatten()

        if isinstance(self.fundamental_matrix, np.ndarray):
            # Convert the truncated RANSAC mask back into a full length mask
            mask[mask] = fmask
            # Set the initial state of the fundamental mask in the masks
            self.masks[maskname] = mask

    def compute_fundamental_error(self, method='equality', clean_keys=[]):
        """
        Given a fundamental matrix, compute the reprojective error between
        a two sets of keypoints.

        Parameters
        ----------
        clean_keys : list
                     of string keys to masking arrays
                     (created by calling outlier detection)

        Returns
        -------
        error : pd.Series
                of reprojective error indexed to the matches data frame
        """
        if self.fundamental_matrix is None:
            log.error('No fundamental matrix has been compute for this edge.')

        matches, mask = self.clean(clean_keys)
        s_keypoints, d_keypoints = self.get_match_coordinates(clean_keys=clean_keys)
        if method == 'equality':
            error = fm.compute_fundamental_error(self.fundamental_matrix, s_keypoints, d_keypoints)
        elif method == 'projection':
            error = fm.compute_reprojection_error(self.fundamental_matrix, s_keypoints, d_keypoints)

        self.costs.loc[mask, 'fundamental_{}'.format(method)] = error

    def compute_homography(self, method='ransac', clean_keys=[], pid=None, maskname='homography', **kwargs):
        """
        For each edge in the (sub) graph, compute the homography.

        Parameters
        ----------
        outlier_algorithm : object
                            An openCV outlier detections algorithm, e.g. cv2.RANSAC

        clean_keys : list
                     of string keys to masking arrays
                     (created by calling outlier detection)
        Returns
        -------
        transformation_matrix : ndarray
                                The 3x3 transformation matrix

        mask : ndarray
               Boolean array of the outliers
        """
        matches, mask = self.clean(clean_keys)

        s_keypoints = self.source.get_keypoint_coordinates(index=matches['source_idx'])
        d_keypoints = self.destination.get_keypoint_coordinates(index=matches['destination_idx'])

        self['homography'], hmask = hm.compute_homography(s_keypoints.values, d_keypoints.values, method=method)

        # Convert the truncated RANSAC mask back into a full length mask
        mask[mask] = hmask.ravel()
        self.masks['homography'] = mask

    def _prep_subpixel(self, nmatches, nstrengths=2):
        """
        Setup the data strutures to return for subpixel matching.

        Parameters
        ----------
        nmatches : int
                    The number of pixels to be subpixel matches

        nstrengths : int
                        The number of 'strength' values to be returned
                        by the subpixel matching method.

        Returns
        -------
        shifts_x : ndarray
                (nmatches, 1) to store the x_shift parameter

        shifts_y : ndarray
                (nmatches, 1) to store the y_shift parameter

        strengths : ndarray
                    (nmatches, nstrengths) to store the strengths for each point

        new_x : ndarray
                (nmatches, 1) to store the updated x coordinates

        new_y : ndarray
                (nmatches, 1) to store the updated y coordinates
        """
        # Setup to store output to append to dataframes
        shifts_x = np.zeros(nmatches)
        shifts_y = np.zeros(nmatches)
        strengths = np.zeros((nmatches, nstrengths))

        new_x = np.empty(nmatches)
        new_y = np.empty(nmatches)

        return shifts_x, shifts_y, strengths, new_x, new_y

    def subpixel_register(self, method='phase', clean_keys=[],
                          template_size=251, search_size=251, **kwargs):
        """
        For the entire graph, compute the subpixel offsets using pattern-matching and add the result
        as an attribute to each edge of the graph.

        Parameters
        ----------
        clean_keys : list
                     of string keys to masking arrays
                     (created by calling outlier detection)

        threshold : float
                    On the range [-1, 1].  Values less than or equal to
                    this threshold are masked and can be considered
                    outliers

        upsampling : int
                     The multiplier to the template and search shapes to upsample
                     for subpixel accuracy

        template_size : int
                        The size of the template in pixels, must be odd. If using phase,
                        only the template size is used.

        search_size : int
                      The size of the search area. When method='template', this size should
                      be >= the template size

        """
        # Build up a composite mask from all of the user specified masks
        matches, mask = self.clean(clean_keys)

        # Get the img handles
        s_img = self.source.geodata
        d_img = self.destination.geodata

        # Determine which algorithm is going ot be used.
        if method == 'phase':
            func = sp.iterative_phase
            nstrengths = 2
        elif method == 'template':
            func = sp.subpixel_template
            nstrengths = 1
        shifts_x, shifts_y, strengths, new_x, new_y = self._prep_subpixel(len(matches), nstrengths)

        # for each edge, calculate this for each keypoint pair
        for i, (idx, row) in enumerate(matches.iterrows()):
            s_idx = int(row['source_idx'])
            d_idx = int(row['destination_idx'])

            if 'source_x' in row.index:
                sx = row.source_x
                sy = row.source_y
            else:
                s_keypoint = self.source.get_keypoint_coordinates([s_idx])
                sx = s_keypoint.x
                sy = s_keypoint.y

            if 'destination_x' in row.index:
                dx = row.destination_x
                dy = row.destination_y
            else:
                d_keypoint = self.destination.get_keypoint_coordinates([d_idx])
                dx = d_keypoint.x
                dy = d_keypoint.y

            if method == 'phase':
                res = sp.iterative_phase(sx, sy, dx, dy, s_img, d_img, size=template_size, **kwargs)
                if res[0]:
                    new_x[i] = res[0]
                    new_y[i] = res[1]
                    strengths[i] = res[2]
            elif method == 'template':
                new_x[i], new_y[i], strengths[i], _ = sp.subpixel_template(sx, sy, dx, dy, s_img, d_img,
                                                                     search_size=search_size,
                                                                     template_size=template_size, **kwargs)

            # Capture the shifts
            shifts_x[i] = new_x[i] - dx
            shifts_y[i] = new_y[i] - dy

        self.matches.loc[mask, 'shift_x'] = shifts_x
        self.matches.loc[mask, 'shift_y'] = shifts_y
        self.matches.loc[mask, 'destination_x'] = new_x
        self.matches.loc[mask, 'destination_y'] = new_y

        if method == 'phase':
            self.costs.loc[mask, 'phase_diff'] = strengths[:,0]
            self.costs.loc[mask, 'rmse'] = strengths[:,1]
        elif method == 'template':
            self.costs.loc[mask, 'correlation'] = strengths[:,0]


    def suppress(self, suppression_func=spf.correlation, clean_keys=[], maskname='suppression', **kwargs):
        """
        Apply a disc based suppression algorithm to get a good spatial
        distribution of high quality points, where the user defines some
        function to be used as the quality metric.

        Parameters
        ----------
        suppression_func : object
                           A function that returns a scalar value to be used
                           as the strength of a given row in the matches data
                           frame.

        suppression_args : tuple
                           Arguments to be passed on to the suppression function

        clean_keys : list
                     of mask keys to be used to reduce the total size
                     of the matches dataframe.
        """
        if not isinstance(self.matches, pd.DataFrame):
            raise AttributeError('This edge does not yet have any matches computed.')

        matches, mask = self.clean(clean_keys)
        rs = self.source.geodata.raster_size
        domain = [0, 0, rs[0], rs[1]]
        # Massage the dataframe into the correct structure
        coords = self.source.get_keypoint_coordinates()
        merged = matches.merge(coords, left_on=['source_idx'], right_index=True)
        merged['strength'] = merged.apply(suppression_func, axis=1, args=([self]))

        smask, k = od.spatial_suppression(merged, domain, **kwargs)

        mask[mask] = smask
        self.masks[maskname] = mask

    def plot_source(self, ax=None, clean_keys=[], **kwargs):  # pragma: no cover
        matches, mask = self.clean(clean_keys=clean_keys)
        indices = pd.Index(matches['source_idx'].values)
        return plot_node(self.source, index_mask=indices, **kwargs)

    def plot_matches(self, clean_keys=[], **kwargs):  # pragme: no cover
        matches, mask = self.clean(clean_keys=clean_keys)
        sourcegd = self.source.geodata
        destingd = self.destination.geodata
        return plot_matches(matches, sourcegd, destingd, **kwargs)

    def plot_destination(self, ax=None, clean_keys=[], **kwargs):  # pragma: no cover
        matches, mask = self.clean(clean_keys=clean_keys)
        indices = pd.Index(matches['destination_idx'].values)
        return plot_node(self.destination, index_mask=indices, **kwargs)

    def plot(self, ax=None, clean_keys=[], node=None, **kwargs):  # pragma: no cover
        dest_keys = [0, '0', 'destination', 'd', 'dest']
        source_keys = [1, '1', 'source', 's']

        # If node is not none, plot a single node
        if node in source_keys:
            return self.plot_source(self, clean_keys=clean_keys, **kwargs)

        elif node in dest_keys:
            return self.plot_destination(self, clean_keys=clean_keys, **kwargs)

        # Else, plot the whole edge
        return plot_edge(self, ax=ax, clean_keys=clean_keys, **kwargs)

    def plot_decomposition(self, *args, **kwargs): #pragma: no cover
        return plot_edge_decomposition(self, *args, **kwargs)

    def clean(self, clean_keys):
        """
        Given a list of clean keys compute the mask of valid
        matches

        Parameters
        ----------
        clean_keys : list
                     of columns names (clean keys)

        Returns
        -------
        matches : dataframe
                  A masked view of the matches dataframe

        mask : series
               A boolean series to inflate back to the full match set
        """
        if clean_keys:
            mask = self.masks[clean_keys].all(axis=1)
        else:
            mask = pd.Series(True, self.matches.index)

        m = mask[mask==True]
        return self.matches.loc[m.index], mask

    def overlap(self):
        """
        Acts on an edge and returns the overlap area and percentage of overlap
        between the two images on the edge. Data is returned to the
        weights dictionary
        """
        poly1 = self.source.geodata.footprint
        poly2 = self.destination.geodata.footprint

        overlapinfo = cg.two_poly_overlap(poly1, poly2)

        self.weights = ('overlap_area', overlapinfo[1])
        self.weights = ('overlap_percn', overlapinfo[0])

    def coverage(self, clean_keys = []):
        """
        Acts on the edge given either the source node
        or the destination node and returns the percentage
        of overlap covered by the keypoints. Data for the
        overlap is gathered from the source node of the edge
        resulting in a maximum area difference of 2% when compared
        to the destination.

        Returns
        -------
        total_overlap_percentage : float
                                   returns the overlap area
                                   covered by the keypoints
        """
        matches, mask = self.clean(clean_keys)
        source_array = self.source.get_keypoint_coordinates(index=matches['source_idx']).values

        source_coords = self.source.geodata.latlon_corners
        destination_coords = self.destination.geodata.latlon_corners

        convex_hull = cg.convex_hull(source_array)

        convex_points = [self.source.geodata.pixel_to_latlon(row[0], row[1]) for row in convex_hull.points[convex_hull.vertices]]
        convex_coords = [(x, y) for x, y in convex_points]

        source_poly = utils.array_to_poly(source_coords)
        destination_poly = utils.array_to_poly(destination_coords)
        convex_poly = utils.array_to_poly(convex_coords)

        intersection_area = cg.get_area(source_poly, destination_poly)

        total_overlap_coverage = (convex_poly.GetArea()/intersection_area)

        return total_overlap_coverage

    def compute_weights(self, clean_keys, **kwargs):
        """
        Computes a voronoi diagram for the overlap between two images
        then gets the area of each polygon resulting in a voronoi weight.
        These weights are then appended to the matches dataframe.

        Parameters
        ----------
        clean_keys : list
                     Of strings used to apply masks to omit correspondences

        """
        if not isinstance(self.matches, pd.DataFrame):
            raise AttributeError('Matches have not been computed for this edge')
        voronoi = cg.compute_voronoi(self, clean_keys, **kwargs)
        self.matches = pd.concat([self.matches, voronoi[1]['vor_weights']], axis=1)

    def compute_overlap(self, buffer_dist=0, **kwargs):
        """
        Estimate a source and destination minimum bounding rectangle, in
        pixel space.
        """
        if not isinstance(self.source.geodata, GeoDataset):
            smbr = None
            dmbr = None
        else:
            try:
                self['overlap_latlon_coords'], smbr, dmbr = self.source.geodata.compute_overlap(self.destination.geodata, **kwargs)
                smbr = list(smbr)
                dmbr = list(dmbr)
                for i in range(4):
                    if i % 2:
                        buf = buffer_dist
                    else:
                        buf = -buffer_dist
                    smbr[i] += buf
                    dmbr[i] += buf

            except:
                smbr = self.source.geodata.xy_extent
                dmbr = self.source.geodata.xy_extent
                log.warning("Overlap between {} and {} could not be "
                                "computed.  Using the full image extents".format(self.source['image_name'],
                                                      self.destination['image_name']))
                smbr = [smbr[0][0], smbr[1][0], smbr[0][1], smbr[1][1]]
                dmbr = [dmbr[0][0], dmbr[1][0], dmbr[0][1], dmbr[1][1]]
        self['source_mbr'] = smbr
        self['destin_mbr'] = dmbr

    def get_match_coordinates(self, clean_keys=[]):
        matches = self.get_matches(clean_keys=clean_keys)

        # no matches, return empty dataframe
        if matches.empty:
            return pd.DataFrame(), pd.DataFrame()

        skps = matches[['source_x', 'source_y']].astype(np.float)
        dkps = matches[['destination_x', 'destination_y']].astype(np.float)

        return skps, dkps

    def get_matches(self, clean_keys=[]): # pragma: no cover
        if self.matches.empty:
            return pd.DataFrame()
        #self.add_coordinates_to_matches()
        matches, _ = self.clean(clean_keys=clean_keys)
        skps = matches[['source_x', 'source_y']]
        dkps = matches[['destination_x', 'destination_y']]
        return matches

class NetworkEdge(Edge):

    default_msg = {'sidx':None,
                    'didx':None,
                    'task':None,
                    'param_step':0,
                    'success':False}

    def __init__(self, *args, **kwargs):
        super(NetworkEdge, self).__init__(*args, **kwargs)
        self.job_status = defaultdict(dict)

    def _from_db(self, table_obj):
        """
        Generic database query to pull the row associated with this node
        from an arbitrary table. We assume that the row id matches the node_id.

        Parameters
        ----------
        table_obj : object
                    The declared table class (from db.model)
        """
        with self.parent.session_scope() as session:
            res = session.query(table_obj).\
                   filter(table_obj.source == self.source['node_id']).\
                   filter(table_obj.destination == self.destination['node_id'])
            session.expunge_all()
        return res

    def compute_homography(self, method='ransac', maskname='homography', **kwargs):
        """
        Estimate the homography and reprojective error on this edge of the graph.

        Parameters
        ----------
        method : {'ransac', 'lmeds', 'normal'}
                 The method that will be used when computing the homography.

        maskname : str
                   The column that the mask will be saved under in the masks dataframe.
                   If the column already exists, then the mask in that column will be overwritten.

        kwargs : dict
                 Extra arguments that will be passed to the homography computation.

        See Also
        --------
        autocnet.transformation.transformations.homography.compute_homography
        """
        matches = self.matches

        s_keypoints = matches[["source_x", "source_y"]].values.astype(np.float64)
        d_keypoints = matches[["destination_x", "destination_y"]].values.astype(np.float64)

        self['homography'], hmask = hm.compute_homography(s_keypoints, d_keypoints, method=method, **kwargs)

        self.masks[maskname] = hmask

    def compute_fundamental_matrix(self, method='ransac', maskname='fundamental', **kwargs):
        """
        Estimate the fundamental matrix (F) using the correspondences tagged to this
        edge.


        Parameters
        ----------
        method : {'ransac', 'lmeds', 'normal', '8point'}
                 The method that will be used when computing the homography.

        maskname : str
                   The column that the mask will be saved under in the masks dataframe.
                   If the column already exists, then the mask in that column will be overwritten.

        kwargs : dict
                 Extra arguments that will be passed to the fundamental matrix computation.

        See Also
        --------
        autocnet.transformation.transformations.fundamental_matrix.compute_fundamental_matrix

        """

        matches = self.matches

        s_keypoints = matches[["source_x", "source_y"]].values.astype(np.float64)
        d_keypoints = matches[["destination_x", "destination_y"]].values.astype(np.float64)
        self.fundamental_matrix, fmask = fm.compute_fundamental_matrix(s_keypoints, d_keypoints, method=method, **kwargs)

        self.masks[maskname] = fmask

    @property
    def weights(self):
        with self.parent.session_scope() as session:
            res = session.query(Edges.weights).\
                                            filter(Edges.source == self.source['node_id']).\
                                            filter(Edges.destination == self.destination['node_id']).\
                                            one()[0]
        self._weights = json.loads(res)
        return self._weights

    @weights.setter
    def weights(self, kv):
        key, value = kv
        with self.parent.session_scope() as session:
            res = session.query(Edges).\
                                            filter(Edges.source == self.source['node_id']).\
                                            filter(Edges.destination == self.destination['node_id']).\
                                            one()
            weights = json.loads(res.weights)
            weights[key] = value

            res.weights = json.dumps(weights)

    @property
    def parent(self):
        return getattr(self, '_parent', None)

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    @property
    def masks(self):
        with self.parent.session_scope() as session:
            res = session.query(Edges.masks).\
                                            filter(Edges.source == self.source['node_id']).\
                                            filter(Edges.destination == self.destination['node_id']).\
                                            first()
        try:
            df = pd.DataFrame.from_records(res[0])
            df.index = df.index.map(int)
        except:
            ids = list(map(int, self.matches.index.values))
            df = pd.DataFrame(index=ids)
        df.index.name = 'match_id'
        return DbDataFrame(df, parent=self, name='masks')

    @masks.setter
    def masks(self, v):

        def dict_check(input):
            for k, v in input.items():
                if isinstance(v, dict):
                    dict_check(v)
                elif v is None:
                    continue
                elif np.isnan(v):
                    input[k] = None


        df = pd.DataFrame(v)
        with self.parent.session_scope() as session:
            res = session.query(Edges).\
                                    filter(Edges.source == self.source['node_id']).\
                                    filter(Edges.destination == self.destination['node_id']).first()
            if res:
                as_dict = df.to_dict()
                dict_check(as_dict)
                # Update the masks
                res.masks = as_dict
                session.add(res)

    @property
    def costs(self):
        # these are np.float coming out, sqlalchemy needs ints
        ids = list(map(int, self.matches.index.values))
        with self.parent.session_scope() as session:
            res = session.query(Costs).filter(Costs.match_id.in_(ids)).all()
            #qf = q.filter(Costs.match_id.in_(ids))
            if res:
                # Parse the JSON dicts in the cost field into a full dimension dataframe
                costs = {r.match_id:r._cost for r in res}
                df = pd.DataFrame.from_records(costs).T  # From records is important because from_dict drops rows with empty dicts
            else:
                df = pd.DataFrame(index=ids)
        df.index.name = 'match_id'
        return DbDataFrame(df, parent=self, name='costs')


    @costs.setter
    def costs(self, v):
        to_db_add = []
        # Get the query obj
        with self.parent.session_scope() as session:
            q = session.query(Costs)
            # Need the new instance here to avoid __setattr__ issues
            df = pd.DataFrame(v)
            for idx, row in df.iterrows():
                # Now invert the expanded dict back into a single JSONB column for storage
                res = q.filter(Costs.match_id == idx).first()
                if res:
                    #update the JSON blob
                    costs_new_or_updated = row.to_dict()
                    for k, v in costs_new_or_updated.items():
                        if v is None:
                            continue
                        elif np.isnan(v):
                            v = None
                        res._cost[k] = v
                    sqlalchemy.orm.attributes.flag_modified(res, '_cost')
                    session.add(res)
                else:
                    row = row.to_dict()
                    costs = row.pop('_costs', {})
                    for k, v in row.items():
                        if np.isnan(v):
                            v = None
                        costs[k] = v
                    cost = Costs(match_id=idx, _cost=costs)
                    to_db_add.append(cost)
            if to_db_add:
                session.bulk_save_objects(to_db_add)

    @property
    def matches(self):
        with self.parent.session_scope() as session:
            q = session.query(Matches)
            qf = q.filter(Matches.source == self.source['node_id'],
                          Matches.destination == self.destination['node_id'])
            odf = pd.read_sql(qf.statement, q.session.bind).set_index('id')
            df = pd.DataFrame(odf.values, index=odf.index.values, columns=odf.columns.values)
            df.index.name = 'id'
        return DbDataFrame(df,  parent=self, name='matches')

    @matches.setter
    def matches(self, v):
        to_db_add = []
        to_db_update = []
        df = pd.DataFrame(v)
        df.index.name = v.index.name
        # Get the query obj
        with self.parent.session_scope() as session:
            q = session.query(Matches)
            for idx, row in df.iterrows():
                # Determine if this is an update or the addition of a new row
                if hasattr(row, 'id'):
                    res = q.filter(Matches.id == row.id).first()
                    match_id = row.id
                elif v.index.name == 'id':
                    res = q.filter(Matches.id == row.name).first()
                    match_id = row.name
                else:
                    res = None
                if res:
                    # update
                    mapping = {}
                    mapping['id'] = match_id
                    for index in row.index:
                        row_val = row[index]
                        if isinstance(row_val, (np.int,)):
                            row_val = int(row_val)
                        elif isinstance(row_val, (np.float,)):
                            row_val = float(row_val)
                        # This should be uncommented if the matches
                        # df is refactored to be a geodataframe
                        #elif isinstance(row_val, WKBElement):
                        #    continue
                        mapping[index] = row_val
                    to_db_update.append(mapping)
                else:
                    match = Matches()
                    # Dynamically iterate over the columns and if the match has an
                    # attribute with the column name, set it.
                    for c in df.columns:
                        if hasattr(match, c):
                            setattr(match, c, row[c])
                    to_db_add.append(match)
            if to_db_add:
                session.bulk_save_objects(to_db_add)
            if to_db_update:
                session.bulk_update_mappings(Matches, to_db_update)

    @matches.deleter
    def matches(self):
        with self.parent.session_scope() as session:
            session.query(Matches).filter(Matches.source == self.source['node_id'], Matches.destination == self.destination['node_id']).delete()

    @property
    def ring(self):
        res = self._from_db(Edges).first()
        if res:
            return res.ring
        return

    @ring.setter
    def ring(self, ring):
        # Setters need a single session and so should not make use of the
        # syntax sugar _from_db
        with self.parent.session_scope() as session:
            res = session.query(Edges).\
                   filter(Edges.source == self.source['node_id']).\
                   filter(Edges.destination == self.destination['node_id']).first()
            if res:
                res.ring = ring
            else:
                edge = Edges(source=self.source['node_id'],
                             destination=self.destination['node_id'],
                             ring=ring)
                session.add(edge)

    @property
    def intersection(self):
        if not hasattr(self, '_intersection'):
            s_fp = self.source.footprint
            d_fp = self.destination.footprint
            self._intersection = s_fp.intersection(d_fp)
        return self._intersection

    @property
    def fundamental_matrix(self):
        res = self._from_db(Edges).first()
        if res:
            return np.asarray(res.fundamental)

    @fundamental_matrix.setter
    def fundamental_matrix(self, v):
        with self.parent.session_scope() as session:
            res = session.query(Edges).\
                   filter(Edges.source == self.source['node_id']).\
                   filter(Edges.destination == self.destination['node_id']).first()
            if res:
                res.fundamental = v
            else:
                edge = Edges(source=self.source['node_id'],
                             destination=self.destination['node_id'],
                             fundamental = v)
                session.add(edge)

    def get_overlapping_indices(self, kps):
        lons_og, lats_og, alts = reproject([kps.xm.values, kps.ym.values, kps.zm.values], semi_major, semi_minor, 'geocent', 'latlon')
        lons, lats = og2oc(lons_og, lats_og, semi_major, semi_minor)
        points = [Point(lons[i], lats[i]) for i in range(len(lons))]
        mask = [i for i in range(len(points)) if self.intersection.contains(points[i])]
        return mask

    def measures(self, filters={}):
        with self.parent.session_scope() as session:
            query = session.query(Measures).filter(sqlalchemy.or_(Measures.imageid == self.source['node_id'], Measures.imageid == self.destination['node_id']))
            for attr, value in filters.items():
                query = query.filter(getattr(Measures,attr)==value)
            res = query.all()
            session.expunge_all()
        return res

    def network_to_matches(self, ignore_point=False, ignore_measure=False, rejected_jigsaw=False):
        """
        For the edge, take any points/measures that are in the database and
        convert them into matches on the associated edge.

        Parameters
        ----------
        ignore_point : bool
                       If False (default) only select the points that are
                       not ignored (currently active).

        ignore_measure : bool
                         If False (default) only add the measures that are
                         not ignored (currently active).

        rejected_jigsaw : bool
                          If False (default) add any points that are not
                          set to jigsaw rejected.

        """
        source = self.source['node_id']
        destin = self.destination['node_id']

        if source > destin:
            source, destin = destin, source

        with self.parent.session_scope() as session:
            q = session.query(Points.id,
                      Points.pointtype,
                      Measures.id.label('mid'),
                      Measures.sample,
                      Measures.line,
                      Measures.measuretype,
                      Measures.imageid,
                      Measures.apriorisample,
                      Measures.aprioriline).\
                filter(Points.ignore==ignore_point,
                       Measures.ignore==ignore_measure,
                       Measures.jigreject==rejected_jigsaw,
                       sqlalchemy.or_(Measures.imageid==source,
                                      Measures.imageid==destin)).join(Measures)

            df = pd.read_sql(q.statement, self.parent.engine)
            matches = []
            columns = ['point_id', 'source_measure_id', 'destin_measure_id', 'source',
                       'source_idx', 'destination', 'destination_idx',
                       'lat', 'lon', 'geom',
                       'source_x', 'source_y', 'source_apriori_x', 'source_apriori_y',
                       'destination_x','destination_y', 'destination_apriori_x', 'destination_apriori_y',
                       'shift_x', 'shift_y', 'original_destination_x',
                       'original_destination_y']

        def net2matches(grp, matches, source, destin):
            # Grab the image ids and then get the cartesian product of the ids to know which
            # edges to put the matches onto
            if len(grp) != 2:
                return

            imagea = grp[grp['imageid'] == source].iloc[0]
            imageb = grp[grp['imageid'] == destin].iloc[0]
            match = [int(imagea.id), int(imagea.mid), int(imageb.mid),
                     source, 0, destin, 0,
                     None, None, None,
                     imagea['sample'], imagea['line'], imagea['apriorisample'], imagea['aprioriline'],
                     imageb['sample'], imageb['line'], imageb['apriorisample'], imageb['aprioriline'],
                     None, None, None, None]
            matches.append(match)

        df.groupby('id').apply(net2matches, matches, source, destin)
        self.matches = pd.DataFrame(matches, columns=columns)

    def mask_to_counter(self, mask):
        """
        Take a mask on an edge and convert the mask into a counter where
        the key is the match id and value is 1 (the match is flagged false).

        TODO: Allow the mask to be an iterable (list). The caller of this should
        then worry about normalization as n-mask strings can come in and we
        cannot anticipate how the user might want to normalize the return.

        Parameters
        ----------
        mask : str
               The name of the mask

        Returns
        -------
          : collections.Counter
            With keys equal to the indices of the False matches
            and values equal to one

        """
        mask = self.masks[mask]
        matches_to_disable = mask[mask == False].index


        bad = {}
        with self.parent.session_scope() as session:
            for o in session.query(Matches).filter(Matches.id.in_(matches_to_disable)).all():
                # This can't just set both to False, we loose a ton of good points - a bad point in 1 image is not
                # necessarily bad in all of the other images. Doing it this way assumes that it is...
                bad[o.source_measure_id] = 1
                bad[o.destin_measure_id] = 1
        return Counter(bad)

    def find_IQR_outliers(self,
                          scaling=1.5,
                          filters={'template_metric': 1, 'template_shift': 0},
                          n_tolerance=10):
        """
        Based on the interquartile range, find the measure outliers from line and sample shifts.

        Parameters
        ----------
        scaling: float
                 scaling factor to use on IQR when determining outlier range

        filters: dict
                 filters used on a match's source measure, only source measures which
                 satisfy filter will be used in outlier calculation

        n_tolerance: int
                     minimum number of measures needed to calculate outliers

        Returns
        -------
        measure_ids: list
                     list of measure ids corresponding to outliers

        resultlog:  dict
                    status of finding outlier
        """
        resultlog = []
        currentlog = {'edge': f'({self.source}, {self.destination})',
                      'status': ''}

        ref_measures = self.measures(filters=filters)
        ref_measure_ids = [m.id for m in ref_measures]

        # use matches where the source measure is the reference measure
        matches = self.matches
        new_match_idx = []
        for i, row in matches.iterrows():
            if (row['source_measure_id'] in ref_measure_ids):
                new_match_idx.append(i)
        matches = matches.loc[new_match_idx]

        # check these is enough data to stastically calculate outliers
        if len(new_match_idx) == 0:
            currentlog['status'] = 'no filtered measures in edge.'
            resultlog.append(currentlog)
            return None, resultlog
        elif len(new_match_idx) < n_tolerance:
            currentlog['status'] = f'{len(new_match_idx)} < {n_tolerance} filtered measures are not statistically signigicant.'
            resultlog.append(currentlog)
            return None, resultlog

        # calculate outliers
        sampleShift = matches['destination_x'] - matches['destination_apriori_x']
        lineShift =  matches['destination_y'] - matches['destination_apriori_y']

        sample_q1, sample_q3 = np.percentile(sampleShift,[25,75])
        line_q1, line_q3 = np.percentile(lineShift,[25,75])
        sample_iqr = sample_q3 - sample_q1
        line_iqr = line_q3 - line_q1

        sample_lowerBound = sample_q1 -(scaling * sample_iqr)
        sample_upperBound = sample_q3 +(scaling * sample_iqr)
        line_lowerBound = line_q1 -(scaling * line_iqr)
        line_upperBound = line_q3 +(scaling * line_iqr)

        sample_outlier_matches = sampleShift[(sampleShift <= sample_lowerBound) | (sampleShift >= sample_upperBound)]
        line_outlier_matches = lineShift[(lineShift <= line_lowerBound) | (lineShift >= line_upperBound)]
        outlier_match_idx = np.concatenate((sample_outlier_matches.index.values, line_outlier_matches.index.values))

        measure_ids = matches.loc[outlier_match_idx]['destin_measure_id'].values
        currentlog['status'] = f'{len(measure_ids)} outliers found.'
        resultlog.append(currentlog)

        return measure_ids, resultlog


    def ignore_outliers(self, outlier_method='IQR', **kwargs):
        """
        Find and ignore outlier measures as determined by outlier method


        Parameters
        ----------
        outlier_method: {'IQR',}
                        method used to determine outliers. Options are: 
                          - interquartile range ('IQR') of line/sample shift

        Returns
        -------
        resultlog: dict
                   status of finding and ignoring outliers


        """
        outlier_dict = {'IQR': self.find_IQR_outliers}
        outlier_func = outlier_dict[outlier_method]

        outlier_destination_mids, resultlog = outlier_func(**kwargs)

        if outlier_destination_mids is None:
            return resultlog

        # ignore outlier measures
        with self.parent.session_scope() as session:
            for m in outlier_destination_mids:
                currentlog = {'measure id': m,
                              'status': 'Ignored'}
                session.query(Measures).filter(Measures.id==m).update({'ignore': True})
                resultlog.append(currentlog)

        return resultlog





import numpy as np
import pandas as pd
import pytest

from plio.io import io_hdf

from .. import keypoints

@pytest.fixture
def kd(scope='module'):
    kps = pd.DataFrame(np.random.random((128,3)), columns=['a', 'b', 'c'])
    desc = np.random.random((128,128))

    return kps, desc

def test_read_write_npy(tmpdir, kd):
    kps, desc = kd
    path = tmpdir.join('out.npz')
    keypoints.to_npy(kps, desc, path.strpath)
    reloaded_kps, reloaded_desc = keypoints.from_npy(path.strpath)

    assert reloaded_kps.equals(kps)
    np.testing.assert_array_equal(reloaded_desc, desc)

def test_read_write_hdf(tmpdir, kd):
    kps, desc = kd
    path = tmpdir.join('out.h5')
    keypoints.to_hdf(path.strpath, keypoints=kps, descriptors=desc)
    reloaded_kps, reloaded_desc = keypoints.from_hdf(path.strpath)

    assert reloaded_kps.equals(kps)
    np.testing.assert_array_equal(reloaded_desc, desc)

    reloaded_kps = keypoints.from_hdf(path.strpath, descriptors=False)
    assert reloaded_kps.equals(kps)

    reloaded_desc = keypoints.from_hdf(path.strpath, keypoints=False)
    np.testing.assert_array_equal(reloaded_desc, desc)

def test_read_hdf_with_index(tmpdir, kd):
    kps, desc = kd                                                                 
    path = tmpdir.join('out.h5')                                                   
    keypoints.to_hdf(path.strpath, keypoints=kps, descriptors=desc) 

    reloaded_kps, reloaded_desc = keypoints.from_hdf(path.strpath, index=np.arange(10))
    assert len(reloaded_kps) == len(reloaded_desc) == 10
    np.testing.assert_array_equal(np.arange(10), reloaded_kps.index.values)

def test_read_write_hdf_with_live_file(tmpdir, kd):
    kps, desc = kd
    path = tmpdir.join('live.h5')
    hf = io_hdf.HDFDataset(path.strpath, mode='w')
    keypoints.to_hdf(hf, keypoints=kps, descriptors=desc)
    reloaded_kps, reloaded_desc = keypoints.from_hdf(hf)

    assert reloaded_kps.equals(kps)
    np.testing.assert_array_equal(reloaded_desc, desc)

@pytest.mark.parametrize("filename, outdir, expected",
                         [('foo', None, 'foo_kps.h5'),
                          ('/path/foo', None, '/path/foo_kps.h5'),
                          ('foo', '/path', '/path/foo_kps.h5')
                         ])
def test_create_output_path(filename, outdir, expected):
    assert keypoints.create_output_path(filename, outdir=outdir) == expected



import os
import sys

import numpy as np
import pandas as pd

from autocnet.cg import cg
from osgeo import ogr
from shapely.geometry import Polygon
from unittest.mock import Mock, MagicMock
from plio.io import io_gdal

from autocnet.graph.node import Node
from autocnet.graph.network import CandidateGraph
from autocnet.graph.edge import Edge
from autocnet.utils.utils import array_to_poly

import pytest

@pytest.fixture
def pts():
    seed = np.random.RandomState(12345)
    return seed.rand(25, 2)

@pytest.fixture
def nspoly():
    return Polygon([(0,0),(.2,0),(.2,1), (0,1), (0,0)])

def test_area_single(pts):
    total_area = 1.0
    ratio = cg.convex_hull_ratio(pts, total_area)

    assert pytest.approx(0.7566490, 5) == ratio

def test_overlap():
    wkt1 = "POLYGON ((0 40, 40 40, 40 0, 0 0, 0 40))"
    wkt2 = "POLYGON ((20 60, 60 60, 60 20, 20 20, 20 60))"

    poly1 = ogr.CreateGeometryFromWkt(wkt1)
    poly2 = ogr.CreateGeometryFromWkt(wkt2)

    info = cg.two_poly_overlap(poly1, poly2)

    assert info[1] == 400
    assert pytest.approx(info[0]) == 14.285714285

def test_geom_mask():
    my_gdf = pd.DataFrame(columns=['x', 'y'], data=[(0, 0), (2, 2)])
    my_poly = Polygon([(1, 1), (3, 1), (3, 3), (1, 3)])
    mask = cg.geom_mask(my_gdf, my_poly)
    assert mask[0] == False
    assert mask[1] == True

@pytest.fixture
def keypoints():
    return pd.DataFrame({'x': (15, 18, 18, 12, 12), 'y': (6, 10, 15, 15, 10)})

def test_voronoi_keypoints(keypoints):
    voronoi_gdf = cg.compute_voronoi(keypoints)
    for i, v in enumerate([12.0, 13.5, 7.5, 7.5, 13.5]):
        assert pytest.approx(voronoi_gdf.weight[i]) == v

def test_voronoi_keypoints_with_geom(keypoints):
    voronoi_gdf = cg.compute_voronoi(keypoints, geometry=True)
    for i, v in enumerate([12.0, 13.5, 7.5, 7.5, 13.5]):
        assert pytest.approx(voronoi_gdf.geometry.area[i]) == v

def test_voronoi_keypoint_intersection(keypoints):
    intersection = Polygon([(10, 5), (20, 5), (20, 20), (10, 20)])
    voronoi_inter_gdf = cg.compute_voronoi(keypoints, intersection)
    for i, v in enumerate([22.5, 26.25, 37.5, 37.5, 26.25]):
        assert pytest.approx(voronoi_inter_gdf.weight[i]) == v

@pytest.mark.parametrize("polygon, nexpected",[
    (Polygon([(0,0), (.2,0), (.2,1), (0,1), (0,0)]), 10),
    (Polygon([(0,0), (1,0), (1,.2), (0,.2), (0,0)]), 10),
    (Polygon([(0,0), (.2, .1), (.2,1.1), (-0.1, 1), (0,0)]), 22)
],
    ids=['vertical', 'horizontal', 'verticalskewed'])
def test_points_in_geom(polygon, nexpected):
    pts = cg.distribute_points_in_geom(polygon)
    assert len(pts) == nexpected
    



import os
from setuptools import setup, find_packages

__version__ = '0.7.2'

def setup_package():
    setup(
        name = "autocnet",
        version = __version__,
        author = "Jay Laura",
        author_email = "jlaura@usgs.gov",
        description = "Automated control network generation.",
        long_description = "Automated sparse control network generation to support photogrammetric control of planetary image data.",
        license = "Public Domain",
        keywords = "Multi-image correspondence detection",
        url = "http://packages.python.org/autocnet",
        packages=find_packages(),
        include_package_data=True,
        zip_safe=False,
        install_requires=[
          "csmapi",
          "cython",
          "cyvlfeat",
          "matplotlib",
          "dill",
          "gdal",
          "geoalchemy2",
          "geopandas",
          "gdal",
          "hoggorm",
          "imageio",
          "kalasiris",
          "knoten ==0.2.0",
          "ncurses",
          "networkx ==2",
          "numpy",
          "opencv <=3.5",
          "plio >=1.3",
          "pandas",
          "pyyaml",
          "plurmy",
          "psycopg2",
          "pvl ==1.0",
          "pyproj",
          "richdem",
          "scikit-image ==0.17",
          "scikit-learn",
          "scipy ==1.2.1",
          "shapely",
          "sqlalchemy ==1.4",
          "sqlalchemy-utils ==0.37.0",
          "redis-py",
          "usgscsm",
          "vlfeat",
          "pip",
          "protobuf",
        ],
        tests_requires=[
            "coveralls",
            "fakeredis",
            "pytest",
            "pytest-cov",
            "pytest-mock"
        ],
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Topic :: Utilities",
            "License :: Public Domain",
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
        ],
        entry_points={"console_scripts": [
        "acn_submit = autocnet.graph.cluster_submit:main",
        "acn_submit_single = autocnet.graph.cluster_submit_single:main"], 
        }
    )

if __name__ == '__main__':
    setup_package()



"""This module contains wrappers around the ISIS campt and mappt that are
specific to understanding the relationship between pixels and projected
coordinates."""

# This is free and unencumbered software released into the public domain.
#
# The authors of autocnet do not claim copyright on the contents of this file.
# For more details about the LICENSE terms and the AUTHORS, you will
# find files of those names at the top level of this repository.
#
# SPDX-License-Identifier: CC0-1.0

import os
from collections import abc
from numbers import Number

import numpy as np

try:
    import kalasiris as isis
except Exception as exception:
    from autocnet.utils.utils import FailedImport
    isis = FailedImport(exception)

import pvl

isis2np_types = {
        "UnsignedByte" : "uint8",
        "SignedWord" : "int16",
        "Double" : "float64",
        "Real" : "float32"
}

np2isis_types = {v: k for k, v in isis2np_types.items()}


def get_isis_special_pixels(arr):
    """
    Returns coordinates of any ISIS no data pixels. Essentially, 
    np.argwhere results of where pixels match ISIS special 
    data types (NIRs, NHRs, HIS, HRS, NULLS).

    Parameters
    ----------
    arr : np.array 
          Array to find special pixels in 
    
    Returns
    -------
    : sp
      np.array of coordinates in y,x format containing special pixel coordinates

    """
    isis_dtype = np2isis_types[str(arr.dtype)]
    sp_pixels = getattr(isis.specialpixels, isis_dtype)

    null = np.argwhere(arr==sp_pixels.Null)
    lrs = np.argwhere(arr==sp_pixels.Lrs)
    lis = np.argwhere(arr==sp_pixels.Lis)
    his = np.argwhere(arr==sp_pixels.His)
    hrs = np.argwhere(arr==sp_pixels.Hrs)
    sp = np.concatenate((null, lrs, lis, his, hrs))

    return sp


def get_nodata_bounds(arr):
    """
    Get bounds for an image that does not contain any ISIS special pixels. That is,
    ISIS Nulls, NIRS, NRS, HIS and HRS pixels

    Parameters
    ----------
    arr : np.array
          2D array representing the image 

    Returns
    -------
    : left_x 
      left x coordinate of new bounds 

    : right_x
      right x coordinate of new bounds 
    
    : top_y
      top y coordinate of new bounds 
    
    : bottom _y
      bottom y coordinates of new bounds 
    """
    sp = get_isis_special_pixels(arr)
    
    if not sp.any():
        return 0, arr.shape[1], 0, arr.shape[0]
    
    cy, cx = arr.shape[1]//2, arr.shape[0]//2
    tree = KDTree(sp, metric='euclidean')

    # For finding K neighbors of P1 with shape (1, 3)
    distances, indices = tree.query(np.array([cy, cx]).reshape(1,2), 1)
    
    # these are slightly misshapen by being in nested arrays (e.g. [[n]], [[y,x]])
    nearest_idx = indices.reshape(1,)
    neary, nearx = sp[nearest_idx].reshape(2,)

    # subtract 1 to exclude the special pixel
    x_dist = abs(cx - nearx) - 1
    y_dist = abs(cy - neary) - 1

    # left_x, right_x, top_y, bottom_y
    left_x = cx - x_dist
    right_x = cx + x_dist
    top_y = cy - y_dist
    bottom_y = cy + y_dist

    return left_x, right_x, top_y, bottom_y


def point_info(
        cube_path: os.PathLike,
        x,
        y,
        point_type: str,
        allowoutside=False
):
    """
    Returns a pvl.collections.MutableMappingSequence object or a
    Sequence of MutableMappingSequence objects which contain keys
    and values derived from the output of ISIS campt or mappt on
    the *cube_path*.

    If x and y are single numbers, then a single MutableMappingSequence
    object will be returned.  If they are Sequences or Numpy arrays, then a
    Sequence of MutableMappingSequence objects will be returned,
    such that the first MutableMappingSequence object of the returned
    Sequence will correspond to the result of *x[0]* and *y[0]*,
    etc.

    Raises subprocess.CalledProcessError if campt or mappt have failures.
    May raise ValueError if campt completes, but reports errors.

    Parameters
    ----------
    cube_path : os.PathLike
                Path to the input cube.

    x : Number, Sequence of Numbers, or Numpy Array
        Point(s) in the x direction. Interpreted as either a sample
        or a longitude value determined by *point_type*.

    y : Number, Sequence of Numbers, or Numpy Array
        Point(s) in the y direction. Interpreted as either a line
        or a latitude value determined by *point_type*.

    point_type : str
                 Options: {"image", "ground"}
                 Pass "image" if  x,y are in image space (sample, line) or
                 "ground" if in ground space (longitude, latitude)

    allowoutside: bool
                  Defaults to False, this parameter is passed to campt
                  or mappt.  Please read the ISIS documentation to
                  learn more about this parameter.

    """
    point_type = point_type.casefold()
    valid_types = {"image", "ground"}
    if point_type not in valid_types:
        raise ValueError(
            f'{point_type} is not a valid point type, valid types are '
            f'{valid_types}'
        )

    if isinstance(x, abc.Sequence) and isinstance(y, abc.Sequence):
        if len(x) != len(y):
            raise IndexError(
                f"Sequences given to x and y must be of the same length."
            )
        x_coords = x
        y_coords = y
    elif isinstance(x, Number) and isinstance(y, Number):
        x_coords = [x, ]
        y_coords = [y, ]
    elif isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
        if not all((x.ndim == 1, y.ndim == 1)):
            raise IndexError(
                f"If they are numpy arrays, x and y must be one-dimensional, "
                f"they were: {x.ndim} and {y.ndim}"
            )
        if x.shape != y.shape:
            raise IndexError(
                f"Numpy arrays given to x and y must be of the same shape."
            )
        x_coords = x
        y_coords = y
    else:
        raise TypeError(
            f"The values of x and y were neither Sequences nor individual "
            f"numbers, they were: {x} and {y}"
        )

    results = []
    if pvl.load(cube_path).get("IsisCube").get("Mapping"):
        # We have a projected image, and must use mappt
        mappt_common_args = dict(allowoutside=allowoutside, type=point_type)

        for xx, yy in zip(x_coords, y_coords):
            mappt_args = {
                "ground": dict(
                    longitude=xx,
                    latitude=yy,
                    coordsys="UNIVERSAL"
                ),
                "image": dict(
                    # Convert PLIO pixels to ISIS pixels
                    sample=xx+0.5,
                    line=yy+0.5
                )
            }
            for k in mappt_args.keys():
                mappt_args[k].update(mappt_common_args)
            mapres = pvl.loads(isis.mappt(cube_path, **mappt_args[point_type]).stdout)["Results"]
            
            # convert from ISIS pixels to PLIO pixels
            mapres['Sample'] = mapres['Sample'] - 0.5
            mapres['Line'] = mapres['Line'] - 0.5

            results.append(mapres)
    else:
        # Not projected, use campt
        if point_type == "ground":
            # campt uses lat, lon for ground but sample, line for image.
            # So swap x,y for ground-to-image calls
            p_list = [f"{lat}, {lon}" for lon, lat in zip(x_coords, y_coords)]
        else:
            p_list = [
                f"{samp+0.5}, {line+0.5}" for samp, line in zip(x_coords, y_coords)
            ]

        # ISIS's campt needs points in a file
        with isis.fromlist.temp(p_list) as f:
            cp = isis.campt(
                cube_path,
                coordlist=f,
                allowoutside=allowoutside,
                usecoordlist=True,
                coordtype=point_type
            )

        camres = pvl.loads(cp.stdout)
        if 'campt' in camres.keys():
            camres = camres['campt']
        for r in camres.getall("GroundPoint"):
            if r['Error'] is None:
                # convert all pixels to PLIO pixels from ISIS
                r["Sample"] -= .5
                r["Line"] -= .5
                
                results.append(r)
            else:
                r["Sample"] = None
                r["Line"] = None
                results.append(r)

    if isinstance(x, (abc.Sequence, np.ndarray)):
        return results
    else:
        return results[0]


def image_to_ground(
        cube_path: os.PathLike,
        sample,
        line,
        lontype="PositiveEast360Longitude",
        lattype="PlanetocentricLatitude",
        allowoutside=False
):
    """
    Returns a two-tuple of numpy arrays or a two-tuple of floats, where
    the first element of the tuple is the longitude(s) and the second
    element are the latitude(s) that represent the coordinate(s) of the
    input *sample* and *line* in *cube_path*.

    If *sample* and *line* are single numbers, then the returned two-tuple
    will have single elements. If they are Sequences, then the returned
    two-tuple will contain numpy arrays.

    Raises the same exceptions as point_info().

    Parameters
    ----------
    cube_path : os.PathLike
                Path to the input cube.

    sample : Number or Sequence of Numbers
        Sample coordinate(s).

    line : Number or Sequence of Numbers
        Line coordinate(s).

    lontype: str
        Name of key to query in the campt or mappt return to get the returned
        longitudes. Defaults to "PositiveEast360Longitude", but other values
        are possible. Please see the campt or mappt documentation.

    lattype: str
        Name of key to query in the campt or mappt return to get the returned
        latitudes. Defaults to "PlanetocentricLatitude", but other values
        are possible.  Please see the campt or mappt documentation.

    """
    res = point_info(cube_path, sample, line, "image", allowoutside=allowoutside)

    if isinstance(sample, (abc.Sequence, np.ndarray)):
        lon_list = list()
        lat_list = list()
        for r in res:
            lon_list.append(_get_value(r[lontype]))
            lat_list.append(_get_value(r[lattype]))

        lons = np.asarray(lon_list)
        lats = np.asarray(lat_list)
    else:
        lons = _get_value(res[lontype])
        lats = _get_value(res[lattype])

    return lons, lats


def _get_value(obj):
    """Returns *obj*, unless *obj* is of type pvl.collections.Quantity, in
    which case, the .value component of the object is returned."""
    if isinstance(obj, pvl.collections.Quantity):
        return obj.value
    else:
        return obj


def ground_to_image(cube_path, lon, lat, allowoutside=False):
    """
    Returns a two-tuple of numpy arrays or a two-tuple of floats, where
    the first element of the tuple is the sample(s) and the second
    element are the lines(s) that represent the coordinate(s) of the
    input *lon* and *lat* in *cube_path*.

    If *lon* and *lat* are single numbers, then the returned two-tuple
    will have single elements. If they are Sequences, then the returned
    two-tuple will contain numpy arrays.

    Raises the same exceptions as point_info().

    Parameters
    ----------
    cube_path : os.PathLike
                Path to the input cube.

    lon: Number or Sequence of Numbers
        Longitude coordinate(s).

    lat: Number or Sequence of Numbers
        Latitude coordinate(s).

    """
    res = point_info(cube_path, lon, lat, "ground", allowoutside=allowoutside)

    if isinstance(lon, (abc.Sequence, np.ndarray)):
        samples, lines = np.asarray([[r["Sample"], r["Line"]] for r in res]).T
    else:
        samples, lines = res["Sample"], res["Line"]

    return samples, lines





import copy
import os
import logging
import time
import sys

import pandas as pd
import pytest

from unittest.mock import patch, PropertyMock, MagicMock

import geopandas as gpd
import numpy as np
from osgeo import ogr
from plio.io import io_gdal

from autocnet.examples import get_path

from .. import network
from .. import edge
from .. import node
import warnings

sys.path.insert(0, os.path.abspath('..'))

LOGGER = logging.getLogger(__name__)

@pytest.fixture()
def graph():
    basepath = get_path('Apollo15')
    return network.CandidateGraph.from_adjacency(get_path('three_image_adjacency.json'),
                                                      basepath=basepath)

@pytest.fixture()
def geo_graph():
    basepath = get_path('Apollo15')
    a = 'AS15-M-0297_crop.cub'
    b = 'AS15-M-0298_crop.cub'
    c = 'AS15-M-0299_crop.cub'
    adjacency = {a:[b,c],
                 b:[a,c],
                 c:[a,b]}
    return network.CandidateGraph.from_adjacency(adjacency, basepath=basepath)

@pytest.fixture()
def reduced_geo():
    basepath = get_path('Apollo15')
    a = 'AS15-M-0297_crop.cub'
    b = 'AS15-M-0298_crop.cub'
    adjacency = {a:[b], b:[a]}
    return network.CandidateGraph.from_adjacency(adjacency, basepath=basepath)

@pytest.fixture()
def disconnected_graph():
    return network.CandidateGraph.from_adjacency(get_path('adjacency.json'))

@pytest.fixture()
def candidategraph(node_a, node_b, node_c):
    # TODO: Getting this fixture from the global conf is causing deepycopy
    # to fail.  Why?
    cg = network.CandidateGraph()

    # Create a candidategraph object - we instantiate a real CandidateGraph to
    # have access of networkx functionality we do not want to test and then
    # mock all autocnet functionality to control test behavior.
    edges = [(0,1,{'data':edge.Edge(0,1)}),
             (0,2,{'data':edge.Edge(0,2)}),
             (1,2,{'data':edge.Edge(1,2)})]

    cg.add_edges_from(edges)



    match_indices = [([0,1,2,3,4,5,6,7], [0,1,2,3,4,5,6,7]),
                     ([0,1,2,3,4,5,8,9], [0,1,2,3,4,5,8,9]),
                     ([0,1,2,3,4,5,8,9], [0,1,2,3,4,5,6,7])]

    matches = []
    for i, e in enumerate(edges):
        c = match_indices[i]
        source_image = np.repeat(e[0], 8)
        destin_image = np.repeat(e[1], 8)
        coords = np.zeros(8)
        data = np.vstack((source_image, c[0], destin_image, c[1],
                          coords, coords, coords, coords)).T
        matches_df = pd.DataFrame(data, columns=['source_image', 'source_idx', 'destination_image', 'destination_idx',
                                                 'source_x', 'source_y', 'destination_x', 'destination_y'])
        matches.append(matches_df)

    # Mock in autocnet methods
    cg.get_matches = MagicMock(return_value=matches)

    # Mock in the node objects onto the candidate graph
    cg.nodes[0]['data'] = node_a
    cg.nodes[1]['data'] = node_b
    cg.nodes[2]['data'] = node_c

    return cg

def test_get_name(graph):
    node_number = graph.graph['node_name_map']['AS15-M-0297_SML.png']
    name = graph.get_name(node_number)
    assert name == 'AS15-M-0297_SML.png'

def test_size(graph):
    assert graph.size() == graph.number_of_edges()
    for u, v, e in graph.edges.data('data'):
        e['edge_weight'] = 10

    assert graph.size('edge_weight') == graph.number_of_edges()*10

def test_serials(geo_graph):
    serials = ['APOLLO15/METRIC/1971-07-31T01:25:02.243',
               'APOLLO15/METRIC/1971-07-31T01:25:27.457',
               'APOLLO15/METRIC/1971-07-31T01:25:52.669']
    for s in serials:
        assert s in geo_graph.serials().values()

def test_unique_fully_connected():
    G = network.CandidateGraph()
    G.add_edges_from([('A', 'B'), ('A', 'C'), ('B', 'C'), ('B', 'D'), ('A', 'E'), ('A', 'F'), ('E', 'F') ])
    fc = G.compute_fully_connected_components()

def test_from_adjacency():
    basepath = get_path('Apollo15')
    a = 'AS15-M-0297_crop.cub'
    b = 'AS15-M-0298_crop.cub'
    c = 'AS15-M-0299_crop.cub'
    adjacency = {a:[b,c],
                 b:[a,c],
                 c:[a,b]}
    g =  network.CandidateGraph.from_adjacency(adjacency, basepath=basepath)
    assert len(g.nodes) == 3
    assert len(g.edges) == 3

    for s, d, e in g.edges.data('data'):
        assert isinstance(e, edge.Edge)
        assert isinstance(g.nodes[s]['data'], node.Node)

def test_add_node(reduced_geo):
    # Test without "image_name" arg (networkx parent method)
    c = 'AS15-M-0299_crop.cub'
    basepath = get_path('Apollo15')
    reduced_geo.add_node(3, data=node.Node(image_name=c,
                                    image_path=os.path.join(basepath, c),
                                    node_id=3))
    assert len(reduced_geo.nodes) == 3
    assert reduced_geo.nodes[3]["data"]["image_name"] == c

def test_add_node_by_name(reduced_geo):
    # Test with "image_name" (cg method)
    c = 'AS15-M-0299_crop.cub'
    basepath = get_path('Apollo15')
    reduced_geo.add_node(image_name=c, basepath=basepath)
    assert len(reduced_geo.nodes) == 3
    assert reduced_geo.nodes[3]["data"]["image_name"] == c
    assert reduced_geo.nodes[1].keys() == reduced_geo.nodes[2].keys() == reduced_geo.nodes[3].keys()

def test_add_node_nonexistent(geo_graph, caplog):
    # Test when "image_name" not found
    with caplog.at_level(logging.WARNING):
        geo_graph.add_node(image_name="nonexistent.jpg")
        assert 'Cannot find nonexistent.jpg' in caplog.text

def test_add_edge():
    basepath = get_path('Apollo15')
    a = 'AS15-M-0297_crop.cub'
    b = 'AS15-M-0298_crop.cub'
    c = 'AS15-M-0299_crop.cub'
    adjacency = {a:[b],
                 b:[a]}
    c_adj = ['AS15-M-0297_crop.cub', 'AS15-M-0298_crop.cub']
    g =  network.CandidateGraph.from_adjacency(adjacency, basepath=basepath)
    g.add_node(image_name=c, basepath=basepath, adjacency=c_adj)

    assert len(g.edges) == 3
    assert g.edges[1, 2]["data"].source == g.nodes[1]["data"]
    assert g.edges[1, 2]["data"].destination == g.nodes[2]["data"]
    assert g.edges[1, 3]["data"].source == g.nodes[1]["data"]
    assert g.edges[1, 3]["data"].destination == g.nodes[3]["data"]
    assert g.edges[2, 3]["data"].source == g.nodes[2]["data"]
    assert g.edges[2, 3]["data"].destination == g.nodes[3]["data"]
    assert g.edges[1, 2].keys() == g.edges[1, 3].keys() == g.edges[2, 3].keys()

def test_add_edge_missing_img(reduced_geo, caplog):
    c = 'AS15-M-0299_crop.cub'
    basepath = get_path('Apollo15')
    # Test when "image_name" not found
    with caplog.at_level(logging.WARNING):
        reduced_geo.add_node(image_name=c, basepath=basepath, adjacency=["nonexistent.jpg"])
        assert 'nonexistent.jpg not found in the graph' in caplog.text

def test_equal(candidategraph):
    cg = copy.deepcopy(candidategraph)
    assert candidategraph == cg

    cg = copy.deepcopy(candidategraph)
    cg.remove_edge(0,1)
    assert candidategraph != cg

    cg = copy.deepcopy(candidategraph)
    cg.remove_node(0)
    assert candidategraph != cg

    cg = copy.deepcopy(candidategraph)
    cg.nodes[0]['image_name'] = 'foo'
    assert candidategraph != cg

    cg = copy.deepcopy(candidategraph)
    cg.edges[0,1]['data']['fundamental_matrix'] = np.random.random((3,3))
    assert candidategraph != cg

def test_get_matches(candidategraph):
    matches = candidategraph.get_matches()
    assert len(matches) == 3
    assert len(matches[0]) == 8
    assert isinstance(matches[0], pd.DataFrame)

def test_island_nodes(disconnected_graph):
    assert len(list(disconnected_graph.island_nodes())) == 1

def test_triangular_cycles(graph):
    cycles = graph.compute_triangular_cycles()
    # Node order is variable, length is not
    assert len(cycles) == 1

def test_connected_subgraphs(graph, disconnected_graph):
    # Calls all return generators, cast to list for positional comparison
    subgraph_list = list(disconnected_graph.connected_subgraphs())
    assert len(subgraph_list) == 2

    islands = list(disconnected_graph.island_nodes())
    assert islands[0] in subgraph_list[1]

    subgraph_list = list(graph.connected_subgraphs())
    assert len(subgraph_list) == 1


def test_filter(graph):

    test_sub_graph = graph.create_node_subgraph([0, 1])

    for n, d in test_sub_graph.nodes(data='data'):
        d.descriptors = np.arange(10)

    for u, v, e in test_sub_graph.edges(data='data'):
        e.matches = pd.DataFrame(np.arange(9).reshape(3,3))

    filtered_nodes = graph.filter_nodes(lambda node: node.descriptors is not None)
    filtered_edges = graph.filter_edges(lambda edge: edge.matches.empty is not True)

    assert filtered_nodes.number_of_nodes() == test_sub_graph.number_of_nodes()
    assert filtered_edges.number_of_edges() == test_sub_graph.number_of_edges()


def test_subset_graph(graph):
    g = graph
    edge_sub = g.create_edge_subgraph([(1, 3)])
    assert len(edge_sub.nodes()) == 2

    node_sub = g.create_node_subgraph([1, 2])
    assert len(node_sub) == 2


def test_subgraph_from_matches(graph):
    test_sub_graph = graph.create_node_subgraph([1, 2])
    #test_sub_graph.extract_features(extractor_parameters={'nfeatures': 25})
    #test_sub_graph.match(k=2)
    for u, v, k in test_sub_graph.edges(data='data'):
        k.matches = pd.DataFrame([1,2,3])
    sub_graph_from_matches = graph.subgraph_from_matches()

    assert test_sub_graph.edges() == sub_graph_from_matches.edges()


def test_minimum_spanning_tree():
    test_dict = {"0": ["4", "2", "1", "3"],
                 "1": ["0", "3", "2", "6", "5"],
                 "2": ["1", "0", "3", "4", "7"],
                 "3": ["2", "0", "1", "5"],
                 "4": ["2", "0"],
                 "5": ["1", "3"],
                 "6": ["1"],
                 "7": ["2"]}

    graph = network.CandidateGraph.from_adjacency(test_dict)
    mst_graph = graph.minimum_spanning_tree()

    assert sorted(mst_graph.nodes()) == sorted(graph.nodes())
    assert len(mst_graph.edges()) == len(graph.edges())-5

@pytest.mark.filterwarnings('ignore::UserWarning')
def test_fromlist():
    mock_list = ['AS15-M-0295_SML.png', 'AS15-M-0296_SML.png', 'AS15-M-0297_SML.png',
                 'AS15-M-0298_SML.png', 'AS15-M-0299_SML.png', 'AS15-M-0300_SML.png']

    good_poly = ogr.CreateGeometryFromWkt('POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))')
    bad_poly = ogr.CreateGeometryFromWkt('POLYGON ((9999 10, 40 40, 20 40, 10 20, 30 10))')

    with patch('plio.io.io_gdal.GeoDataset.footprint', new_callable=PropertyMock) as patch_fp:
        patch_fp.return_value = good_poly
        n = network.CandidateGraph.from_filelist(mock_list, get_path('Apollo15'))
        assert n.number_of_nodes() == 6
        assert n.number_of_edges() == 15

        patch_fp.return_value = bad_poly
        n = network.CandidateGraph.from_filelist(mock_list, get_path('Apollo15'))
        assert n.number_of_nodes() == 6
        assert n.number_of_edges() == 0

    n = network.CandidateGraph.from_filelist(mock_list, get_path('Apollo15'))
    assert len(n.nodes()) == 6

    n = network.CandidateGraph.from_filelist(get_path('adjacency.lis'), get_path('Apollo15'))
    assert len(n.nodes()) == 6


def test_apply_func_to_edges(graph):

    try:
        graph.apply_func_to_edges('incorrect_func')
    except AttributeError:
        pass

    return_value = pd.DataFrame(np.arange(36).reshape(6,6), columns=['a', 'b', 'c', 'source_idx', 'destination_idx', 'distance'])
    with patch('autocnet.graph.edge.Edge.matches', new_callable=PropertyMock, return_value=return_value):
        graph.apply_func_to_edges("symmetry_check")

        # Test passing the func by signature
        graph.apply_func_to_edges(graph[1][2]['data'].symmetry_check)
        assert not graph[1][3]['data'].masks.symmetry.all()

def test_generate_control_network(candidategraph):
    candidategraph.generate_control_network()
    cn = candidategraph.controlnetwork

    # Confirm that points are being created with measures.
    for i, g in cn.groupby('point_id'):
        assert len(g) == 2

def test_set_maxsize(graph):
    maxsizes = network.MAXSIZE
    assert(graph.maxsize == maxsizes[0])
    graph.maxsize = 12
    assert(graph.maxsize == maxsizes[12])
    with pytest.raises(KeyError):
        graph.maxsize = 7

def test_update_data(graph):
   ctime = graph.graph['modifieddate']
   time.sleep(1)
   graph._update_date()
   ntime = graph.graph['modifieddate']
   assert ctime != ntime

def test_is_complete(graph):
    # Create a small incomplete graph with three nodes and two edges
    incomplete_graph = network.CandidateGraph()
    incomplete_graph.add_nodes_from([1, 2, 3])
    incomplete_graph.add_edges_from([(1, 2), (2, 3)])

    assert False == incomplete_graph.is_complete()
    assert True == graph.is_complete()

def test_get_matches(candidategraph):
    matches = candidategraph.get_matches()
    assert len(matches) == 3
    assert 'source_x' in matches[0].columns
    assert len(matches[0]) == 8

def test_apply(graph):
    def set_matches(e):
        e.matches = pd.DataFrame(['fake', 'fake', 'fake'])

    def get_matches(e):
        return e.matches

    graph.apply(set_matches)
    results = graph.apply(get_matches)

    for matches in results:
        assert len(matches) == 3
    assert len(matches) == len(graph.edges)

def test_apply_on_nodes(graph):
    def set_test_attribute(n):
        n.test_attribute = 1

    def get_test_attribute(n):
        return n.test_attribute

    for _, data in graph.nodes(data=True):
        if data['data']['image_name'] == 'AS15-M-0297_SML.png':
            data['data'].ignore = True

    graph.apply(set_test_attribute, on='node')
    results = graph.apply(get_test_attribute, on='node')

    assert len(results) == len(graph.nodes) - 1
    for test_att in results:
        assert test_att == 1

def test_tofilelist(graph):
    flist = graph.to_filelist()
    print(flist)
    truth = ['AS15-M-0297_SML.png', 'AS15-M-0298_SML.png', 'AS15-M-0299_SML.png']
    basenames = sorted([os.path.basename(i) for i in flist])
    assert truth == basenames

def test_footprints(geo_graph):
    # This is just testing the interface - should get a geodataframe back
    assert isinstance(geo_graph.footprints(), gpd.GeoDataFrame)



import numpy as np
import networkx as nx

from matplotlib import pyplot as plt
import matplotlib

from skimage.transform import resize

def downsample(array, amount):
    return resize(array,
                      (int(array.shape[0] / amount),
                      int(array.shape[1] / amount)))

def plot_graph(graph, ax=None, cmap='Spectral', labels=False, font_size=12, clusters=None, **kwargs):
    """

    Parameters
    ----------
    graph : object
            A networkX or derived graph object

    ax : objext
         A MatPlotLib axes object

    cmap : str
           A MatPlotLib color map string. Default 'Spectral'

    Returns
    -------
    ax : object
         A MatPlotLib axes object. Either the argument passed in
         or a new object
    """
    if ax is None:
        ax = plt.gca()

    cmap = matplotlib.cm.get_cmap(cmap)

    # Setup edge color based on the health metric
    colors = []
    for s, d, e in graph.edges_iter(data=True):
        if hasattr(e, 'health'):
            colors.append(cmap(e.health)[0])
        else:
            colors.append(cmap(0)[0])

    pos = nx.spring_layout(graph)
    nx.draw_networkx_nodes(graph, pos, ax=ax)
    nx.draw_networkx_edges(graph, pos, ax=ax)
    if labels:
        labels = dict((i, d['image_name']) for i, d in graph.nodes_iter(data=True))
        nx.draw_networkx_labels(graph, pos, labels, font_size=font_size, ax=ax)
    ax.axis('off')
    return ax


def plot_node(node, ax=None, clean_keys=[], index_mask=None, downsampling=1, **kwargs):
    """
    Plot the array and keypoints for a given node.

    Parameters
    ----------
    node : object
           A Node object from which data is extracted

    ax : object
         A MatPlotLIb axes object

    clean_keys : list
                 of strings of masking array names to apply

    kwargs : dict
             of MatPlotLib plotting options

    Returns
    -------
    ax : object
         A MatPlotLib axes object.  Either the argument passed in
         or a new object
    """

    if ax is None:
        ax = plt.gca()

    band = 1
    if 'band' in kwargs.keys():
        band = kwargs['band']
        kwargs.pop('band', None)

    array = node.get_array(band)

    if isinstance(downsampling, bool):
        downsampling = node['downsample_amount']

    array = downsample(array, downsampling)

    ax.set_title(node['image_name'])
    ax.margins(tight=True)
    ax.axis('off')

    if 'cmap' in kwargs:
        cmap = kwargs['cmap']
    else:
        cmap = 'Greys'

    ax.imshow(array, cmap=cmap)

    keypoints = node.get_keypoints(index=index_mask)

    if not keypoints.empty:
        marker = '.'
        if 'marker' in kwargs.keys():
            marker = kwargs['marker']
            kwargs.pop('marker', None)
        color = 'r'
        if 'color' in kwargs.keys():
            color = kwargs['color']
            kwargs.pop('color', None)
        ax.scatter(keypoints['x'], keypoints['y'], marker=marker, color=color, **kwargs)

    return ax


def plot_edge_decomposition(edge, ax=None, clean_keys=[], image_space=100,
                            scatter_kwargs={}, line_kwargs={}, image_kwargs={}):

    if ax is None:
        ax = plt.gca()

    # Plot setup
    ax.set_title('Matching: {} to {}'.format(edge.source['image_name'],
                                             edge.destination['image_name']))
    ax.margins(tight=True)
    ax.axis('off')

    # Image plotting
    source_array = edge.source.get_array()
    destination_array = edge.destination.get_array()

    s_shape = source_array.shape
    d_shape = destination_array.shape

    y = max(s_shape[0], d_shape[0])
    x = s_shape[1] + d_shape[1] + image_space
    composite = np.zeros((y, x))
    composite_decomp = np.zeros((y, x), dtype=np.int16)

    composite[0: s_shape[0], :s_shape[1]] = source_array
    composite[0: d_shape[0], s_shape[1] + image_space:] = destination_array

    composite_decomp[0: s_shape[0], :s_shape[1]] = edge.smembership
    composite_decomp[0: d_shape[0], s_shape[1] + image_space:] = edge.dmembership

    if 'cmap' in image_kwargs:
        cmap = image_kwargs['cmap']
    else:
        cmap = 'Greys'

    matches, mask = edge.clean(clean_keys)

    source_keypoints = edge.source.get_keypoints(index=matches['source_idx'])
    destination_keypoints = edge.destination.get_keypoints(index=matches['destination_idx'])

    # Plot the source
    source_idx = matches['source_idx'].values
    s_kps = source_keypoints.loc[source_idx]
    ax.scatter(s_kps['x'], s_kps['y'], **scatter_kwargs, cmap='gray')

    # Plot the destination
    destination_idx = matches['destination_idx'].values
    d_kps = destination_keypoints.loc[destination_idx]
    x_offset = s_shape[1] + image_space
    newx = d_kps['x'] + x_offset
    ax.scatter(newx, d_kps['y'], **scatter_kwargs)

    ax.imshow(composite, cmap=cmap)
    ax.imshow(composite_decomp, cmap='spectral', alpha=0.35)
    # Draw the connecting lines
    color = 'y'
    if 'color' in line_kwargs.keys():
        color = line_kwargs['color']
        line_kwargs.pop('color', None)

    s_kps = s_kps[['x', 'y']].values
    d_kps = d_kps[['x', 'y']].values
    d_kps[:, 0] += x_offset

    for l in zip(s_kps, d_kps):
        ax.plot((l[0][0], l[1][0]), (l[0][1], l[1][1]), color=color, **line_kwargs)

    return ax

def plot_edge(edge, ax=None, clean_keys=[], image_space=100, downsampling=1,
              scatter_kwargs={}, line_kwargs={}, image_kwargs={}):
    """
    Plot the correspondences for a given edge

    Parameters
    ----------
    edge : object
           A graph edge object

    ax : object
         A MatPlotLIb axes object

    clean_keys : list
                 of strings of masking array names to apply

    image_space : int
                  The number of pixels to insert between the images

    downsample : bool

    scatter_kwargs : dict
                     of MatPlotLib arguments to be applied to the scatter plots

    line_kwargs : dict
                  of MatPlotLib arguments to be applied to the lines connecting matches

    image_kwargs : dict
                   of MatPlotLib arguments to be applied to the image rendering

    Returns
    -------
    ax : object
         A MatPlotLib axes object.  Either the argument passed in
         or a new object
    """

    if ax is None:
        ax = plt.gca()

    # Plot setup
    ax.set_title('Matching: {} to {}'.format(edge.source['image_name'],
                                             edge.destination['image_name']))
    ax.margins(tight=True)
    ax.axis('off')

    # Image plotting
    if isinstance(downsampling, bool):
        downsample_source = edge.source['downsample_amount']
    else:
        downsample_source = downsampling
    source_array = edge.source.get_array()
    source_array = downsample(source_array, downsample_source)

    if isinstance(downsampling, bool):
        downsample_destin = edge.destination['downsample_amount']
    else:
        downsample_destin = downsampling
    destination_array = edge.destination.get_array()
    destination_array = downsample(destination_array, downsample_destin)

    s_shape = source_array.shape
    d_shape = destination_array.shape

    y = max(s_shape[0], d_shape[0])
    x = s_shape[1] + d_shape[1] + image_space
    composite = np.zeros((y, x))

    composite[0: s_shape[0], :s_shape[1]] = source_array
    composite[0: d_shape[0], s_shape[1] + image_space:] = destination_array

    if 'cmap' in image_kwargs:
        image_cmap = image_kwargs['cmap']
    else:
        image_cmap = 'Greys'

    matches, mask = edge.clean(clean_keys)

    if not matches.empty:
        source_keypoints = edge.source.get_keypoints(index=matches['source_idx'])
        destination_keypoints = edge.destination.get_keypoints(index=matches['destination_idx'])

        # Plot the source
        source_idx = matches['source_idx'].values
        s_kps = source_keypoints.loc[source_idx]
        ax.scatter(s_kps['x'], s_kps['y'], **scatter_kwargs)

        # Plot the destination
        destination_idx = matches['destination_idx'].values
        d_kps = destination_keypoints.loc[destination_idx]
        x_offset = s_shape[1] + image_space
        newx = d_kps['x'] + x_offset
        ax.scatter(newx, d_kps['y'], **scatter_kwargs)

        # Draw the connecting lines
        color = 'y'
        if 'color' in line_kwargs.keys():
            color = line_kwargs['color']
            line_kwargs.pop('color', None)

        s_kps = s_kps[['x', 'y']].values
        d_kps = d_kps[['x', 'y']].values
        d_kps[:, 0] += x_offset

        for l in zip(s_kps, d_kps):
            ax.plot((l[0][0], l[1][0]), (l[0][1], l[1][1]), color=color, **line_kwargs)

    ax.imshow(composite, cmap=image_cmap)

    return ax

def cluster_plot(graph, ax=None, cmap='Spectral'):  # pragma: no cover
    """
    Parameters
    ----------
    graph : object
            A networkX or derived graph object

    ax : object
         A MatPlotLib axes object

    cmap : str
           A MatPlotLib color map string. Default 'Spectral'

    Returns
    -------
    ax : object
         A MatPlotLib axes object that was either passed in
         or a new axes object
    """
    if ax is None:
        ax = plt.gca()

    if not hasattr(graph, 'clusters'):
        raise AttributeError('Clusters have not been computed.')

    cmap = matplotlib.cm.get_cmap(cmap)

    colors = []

    for i, n in graph.nodes_iter(data=True):
        for j in enumerate(graph.clusters):
            if i in graph.clusters.get(j[1]):
                colors.append(cmap(j[1])[0])
                continue

    nx.draw(graph, ax=ax, node_color=colors)
    return ax

def plot_matches(matches, a, b, extent=200):
    nsubplots = len(matches)
    figs = []
    for i, (idx, row) in enumerate(matches.iterrows()):
        fig, axes = plt.subplots(1, 2)
        sx = row.source_x
        sy = row.source_y
        dx = row.destination_x
        dy = row.destination_y

        suba = a.read_array(pixels=[int(sx-extent), int(sy-extent), int(2*extent), int(2*extent)])
        subb = b.read_array(pixels=[int(dx-extent), int(dy-extent), int(2*extent), int(2*extent)])

        axes[0].imshow(suba, cmap='Greys')
        axes[1].imshow(subb, cmap='Greys')
        figs.append(fig)

    return figs



import pytest
import numpy as np

from datetime import datetime
from inspect import signature
import json

from shapely.geometry import Point

from autocnet.utils.serializers import JsonEncoder, object_hook

@pytest.mark.parametrize("data, serialized", [
    ({'foo':np.arange(5)}, {"foo": [0, 1, 2, 3, 4]}),
    ({'foo':np.int64(1)}, {"foo": 1}),
    ({'foo':b'bar'}, {"foo": "bar"}),
    ({'foo':set(['a', 'b', 'c'])}, {"foo": ["a", "b", "c"]}),
    ({'foo':Point(0,0)}, {"foo": 'POINT (0 0)'}),
    ({'foo':datetime(1982, 9, 8)}, {"foo": '1982-09-08 00:00:00'})
])
def test_json_encoder(data, serialized):
    res = json.dumps(data, cls=JsonEncoder)
    res = json.loads(res, object_hook=object_hook)
    if isinstance(res['foo'], list):
        res['foo'] = sorted(res['foo'])
    elif isinstance(res['foo'], Point):
        res['foo'] = res['foo'].wkt
    assert res == serialized

@pytest.mark.parametrize("data", [
    ({'func':lambda x:True}),
    ({'func':lambda x:True, 'other':1}),
    ({'func':lambda x:True, 'other':1,
       'nested':{'inside':'foo'}}),
    ({'func':lambda x,y:x+y})
])
def test_json_roundtrip(data):
    as_str = json.dumps(data, cls=JsonEncoder)
    as_dict = json.loads(as_str, object_hook=object_hook)
    for k, v in data.items():
        if callable(v):
            sig = signature(v)
            nparams = len(sig.parameters)
            args = [True] * nparams
            assert as_dict[k](*args) == v(*args)
            continue
        assert as_dict[k] == v


from math import isclose, ceil
import random
import warnings

import pandas as pd
import numpy as np
import networkx as nx
import geopandas as gpd
import ogr
import logging

from skimage import transform as tf
from scipy.spatial import Voronoi, Delaunay, ConvexHull
import shapely.geometry
from shapely.geometry import Polygon, MultiPolygon, Point
from shapely.affinity import scale
from shapely import wkt, geometry

from autocnet.utils import utils
from autocnet.cg import cg

from shapely.ops import cascaded_union, polygonize

# set up the logger file
log = logging.getLogger(__name__)

def two_point_extrapolate(x, xs, ys):
    """

    Parameters
    ----------
    x : float
             point where you want corresponding y value

    xs : ndarray
             (1, 2) array of point x coordinates

    ys : ndarray
             (1, 2) array of point x coordinates

    Returns
    -------
    y : float
            extrapolated value associated with x

    """

    m = (ys[1]-ys[0])/(xs[1]-xs[0])
    y = ys[0] + m*(x-xs[0])

    return x, y

def convex_hull_ratio(points, ideal_area):
    """

    Parameters
    ----------
    points : ndarray
             (n, 2) array of point coordinates

    ideal_area : float
                 The total area that could be covered

    Returns
    -------
    ratio : float
            The ratio convex hull volume / ideal_area

    """
    hull = ConvexHull(points)
    return hull.volume / ideal_area


def convex_hull(points):

    """

    Parameters
    ----------
    points : ndarray
             (n, 2) array of point coordinates

    Returns
    -------
    hull : 2-D convex hull
            Provides a convex hull that is used
            to determine coverage

    """

    if isinstance(points, pd.DataFrame) :
        points = pd.DataFrame(points).values

    hull = ConvexHull(points)
    return hull


def geom_mask(keypoints, geom): # ADDED
    """
    Masks any points that are outside of the bounds of the given
    geometry.

    Parameters
    ----------
    keypoints : dataframe
                      A pandas dataframe of points to mask

    geom : object
                Shapely geometry object to use as a mask
    """

    def _in_mbr(r, mbr):
        if (mbr[0] <= r.x <= mbr[2]) and (mbr[1] <= r.y <= mbr[3]):
            return True
        else:
            return False

    mbr = geom.bounds
    initial_mask = keypoints.apply(_in_mbr, axis=1, args=(mbr,))

    return initial_mask


def two_poly_overlap(poly1, poly2):
    """

    Parameters
    ----------
    poly1 : ogr polygon
            Any polygon that shares some kind of overlap
            with poly2

    poly2 : ogr polygon
            Any polygon that shares some kind of overlap
            with poly1

    Returns
    -------
    overlap_percn : float
                    The percentage of image overlap

    overlap_area : float
                   The total area of overalap

    """
    overlap_area_polygon = poly2.Intersection(poly1)
    overlap_area = overlap_area_polygon.GetArea()
    area1 = poly1.GetArea()
    area2 = poly2.GetArea()

    overlap_percn = (overlap_area / (area1 + area2 - overlap_area)) * 100
    return overlap_percn, overlap_area, overlap_area_polygon


def get_area(poly1, poly2):
    """

    Parameters
    ----------
    poly1 : ogr polygon
            General ogr polygon

    poly2 : ogr polygon
            General ogr polygon

    Returns
    -------
    intersection_area : float
                        returns the intersection area
                        of two polygons

    """
    intersection_area = poly1.Intersection(poly2).GetArea()
    return intersection_area


def compute_voronoi(keypoints, intersection=None, geometry=False, s=30): # ADDED
        """
        Creates a voronoi diagram for all edges in a graph, and assigns a given
        weight to each edge. This is based around voronoi polygons generated
        by scipy's voronoi method, to determine if an image has significant coverage.

        Parameters
        ----------
        graph : object
               A networkx graph object

        clean_keys : list
                     Of strings used to apply masks to omit correspondences

        s : int
            Offset for the corners of the image
        """
        vor_keypoints = []

        keypoints.apply(lambda x: vor_keypoints.append((x['x'], x['y'])), axis = 1)

        if intersection is None:
            keypoint_bounds = Polygon(vor_keypoints).bounds
            intersection = shapely.geometry.box(keypoint_bounds[0], keypoint_bounds[1],
                                                    keypoint_bounds[2], keypoint_bounds[3])

        scaled_coords = np.array(scale(intersection, s, s).exterior.coords)

        vor_keypoints = np.vstack((vor_keypoints, scaled_coords))
        vor = Voronoi(vor_keypoints)
        # Might move the code below to its own method depending on feedback
        if geometry:
            voronoi_df = gpd.GeoDataFrame(data = keypoints, columns=['x', 'y', 'weight', 'geometry'])
        else:
            voronoi_df = gpd.GeoDataFrame(data = keypoints, columns=['x', 'y', 'weight'])

        i = 0
        vor_points = np.asarray(vor.points)
        for region in vor.regions:
            region_point = vor_points[np.argwhere(vor.point_region==i)]

            if not -1 in region:
                polygon_points = [vor.vertices[i] for i in region]

                if len(polygon_points) != 0:
                    polygon = Polygon(polygon_points)

                    intersection_poly = polygon.intersection(intersection)

                    voronoi_df.loc[(voronoi_df["x"] == region_point[0][0][0]) &
                                   (voronoi_df["y"] == region_point[0][0][1]),
                                   'weight'] = intersection_poly.area
                    if geometry:
                        voronoi_df.loc[(voronoi_df["x"] == region_point[0][0][0]) &
                                       (voronoi_df["y"] == region_point[0][0][1]),
                                       'geometry'] = intersection_poly
            i += 1

        return voronoi_df

def single_centroid(geom):
    """
    For a geom, return the centroid

    Parameters
    ----------
    geom : shapely.geom object

    Returns
    -------

     : list
            in the form [(x,y)]
    """
    x, y = geom.centroid.xy
    return [(x[0],y[0])]

def nearest(pt, search):
    """
    Fine the index of nearest (Euclidean) point in a list
    of points.

    Parameters
    ----------
    pt : ndarray
         (2,1) array
    search : ndarray
             (n,2) array of points to search within. The
             returned index is the closet point in this set
             to the search

    Returns
    -------
     : int
       The index to the nearest point.
    """
    return np.argmin(np.sum((search - pt)**2, axis=1))

def create_points_along_line(p1, p2, npts):
    """
    Compute a set of nodes equally spaced between
    two points, not including the end points.

    Parameters
    ----------
    p1 : iterable
         in the form (x,y)

    p2 : iterable
         in the form(x,y)

    npts : int
           The number of nodes to be returned

    Returns
    -------
     : ndarray
       (n,2) array of nodes
    """
    # npts +2 since the endpoints are included in linspace
    # but this func clips them
    return np.linspace(p1, p2, npts+2)[1:-1]

def xy_in_polygon(x,y, geom):
    """
    Returns true is an x,y pair is contained within
    the geom.

    Parameters
    ----------
    x : Number
        The x coordinate

    y : Number
        The y coordinate

    Returns
    -------
     : bool
       True if the point is contained within the geom.
    """
    return geom.contains(Point(x, y))

def generate_random(number, polygon):
    points = []
    minx, miny, maxx, maxy = polygon.bounds
    i = 0
    while len(points) < number and i < 1000:
        pnt = Point(random.uniform(minx, maxx), random.uniform(miny, maxy))
        if polygon.contains(pnt):
            points.append([pnt.x, pnt.y])
        i += 1
    return np.asarray(points)

def distribute_points_classic(geom, nspts, ewpts, use_mrr=True, **kwargs):
    """
    This is a decision tree that attempts to perform a
    very simplistic approximation of the shape
    of the geometry and then place some number of
    north/south and east/west points into the geometry.

    Parameters
    ----------
    geom : shapely.geom
           A shapely geometry object

    nspts : int
            The number of points to attempt to place
            in the N/S (up/down) direction

    ewpts : int
            The number of points to attempt to place
            in the E/W (right/left) direction
    
    use_mrr : boolean
              If True (default) compute the minimum rotated rectangle bounding
              the geometry

    Returns
    -------
    valid : list
            of point coordinates in the form [(x1,y1), (x2,y2), ..., (xn, yn)]
    """
    original_geom = geom
    if use_mrr:
        geom = geom.minimum_rotated_rectangle
        
    geom_coords = np.column_stack(geom.exterior.xy)
    coords = np.array(list(zip(*geom.envelope.exterior.xy))[:-1])

    ll = coords[0]
    lr = coords[1]
    ur = coords[2]
    ul = coords[3]

    # Find the points nearest the ul and ur
    ul_actual = geom_coords[nearest(ul, geom_coords)]
    ur_actual = geom_coords[nearest(ur, geom_coords)]
    newtop = create_points_along_line(ul_actual, ur_actual, ewpts)

    # Find the points nearest the ll and lr
    ll_actual = geom_coords[nearest(ll, geom_coords)]
    lr_actual = geom_coords[nearest(lr, geom_coords)]
    newbot = create_points_along_line(ll_actual, lr_actual, ewpts)

    points = []
    for i in range(len(newtop)):
        top = newtop[i]
        bot = newbot[i]

        line_of_points = create_points_along_line(top, bot, nspts)
        points.append(line_of_points)

    if len(points) < 1:
        return []

    points = np.vstack(points)
    # Perform a spatial intersection check to eject points that are not valid
    valid = [p for p in points if xy_in_polygon(p[0], p[1], original_geom)]
    # The standard method failed. Attempt random placement within the geometry
    if not valid:
        valid = generate_random(ewpts * nspts, original_geom)
    return valid

def distribute_points_new(geom, nspts, ewpts, Session):
    """
    This is a decision tree that attempts to perform a
    very simplistic approximation of the shape
    of the geometry and then place some number of
    north/south and east/west points into the geometry.
    This function works best on bulky geometries, such as
    a combination of all network image footprints.

    Parameters
    ----------
    geom : shapely.geom
           A shapely geometry object which is a union of
           all of the image footprints in the network
           that is being grounded.

    nspts : int
            The number of points to attempt to place
            in the N/S (up/down) direction

    ewpts : int
            The number of points to attempt to place
            in the E/W (right/left) direction

    Returns
    -------
    valid : list
            of point coordinates in the form [(x1,y1), (x2,y2), ..., (xn, yn)]
    """
    coords = np.array(list(zip(*geom.envelope.exterior.xy))[:-1])

    ll = coords[0]
    lr = coords[1]
    ur = coords[2]
    ul = coords[3]

    rr_coords = np.array(list(zip(*geom.minimum_rotated_rectangle.exterior.xy))[:-1])
    w = min([i[0] for i in rr_coords])
    s = min([i[1] for i in rr_coords])
    swid = nearest([w, s], rr_coords)
    rr_coords = np.vstack([rr_coords[swid:], rr_coords[0:swid]]) # reorder to match envelope/coords order

    x = np.linspace(ul[0], ur[0], ewpts+2)[1:-1]
    y = np.linspace(ul[1], ll[1], nspts+2)[1:-1]

    grid = np.transpose([np.tile(x, len(y)), np.repeat(y, len(x))])

    if len(grid) < 1:
        return []

    affine = tf.estimate_transform('affine', coords, rr_coords)
    rr_grid = affine(grid)

    # Perform a spatial intersection check to eject points that are not valid
    valid = [p for p in rr_grid if xy_in_polygon(p[0], p[1], geom)]
    return valid

def distribute_points_in_geom(geom, method="classic",
                              nspts_func=lambda x: ceil(round(x,1)*10),
                              ewpts_func=lambda x: ceil(round(x,1)*5),
                              Session=None,
                              **kwargs):
    """
    Given a geometry, attempt a basic classification of the shape.
    RIght now, this simply attempts to determine if the bounding box
    is generally N/S or generally E/W trending. Once the determination
    is made, the algorithm places points in the geometry and returns
    a list of valid (intersecting) points.

    The kwargs for this algorithm take a function that expects a number
    as an input and returns an integer number of points to place. The
    input number is the distance between the top/bottom or left/right
    sides of the geometry.

    This algorithm does not know anything about the units being used
    so the caller is responsible for acocunting for units (if appropriate)
    in the passed funcs.

    Parameters
    ----------
    geom : shapely.geom object
           The geometry object

    nspts_func : obj
                 Function taking a Number and returning an int

    ewpts_func : obj
                 Function taking a Number and returning an int

    Returns
    -------
    valid : np.ndarray
            of valid points in the form (x,y) or (lon,lat)

    """

    point_funcs = {
        "classic" :  distribute_points_classic,
        "new" : distribute_points_new
    }

    point_distribution_func = point_funcs[method]

    coords = list(zip(*geom.envelope.exterior.xy))
   

    # This logic is kwarg swapping - need to trace this logic.
    short = np.inf
    long = -np.inf
    shortid = 0
    longid = 0
    for i, p in enumerate(coords[:-1]):
        d = np.sqrt((coords[i+1][0] - p[0])**2+(coords[i+1][1]-p[1])**2)
        if d < short:
            short = d
            shortid = i
        if d > long:
            long = d
            longid = i
    ratio = short/long
    
    ns = False
    ew = False
    valid = []
    
    # The polygons should be encoded with a lower left origin in counter-clockwise direction.
    # Therefore, if the 'bottom' is the short edge it should be id 0 and modulo 2 == 0.
    if shortid % 2 == 0:
        # Also if the geom is a perfect square
        ns = True
    elif longid % 2 == 0:
        ew = True

    # Decision Tree
    if ratio < 0.1 and geom.area < 0.01:
        # Class: Slivers - ignore.
        return np.array([])
    elif geom.area <= 0.004 and ratio >= 0.25:
        # Single point at the centroid
        valid = single_centroid(geom)
    elif ns==True:
        # Class, north/south poly, multi-point
        nspts = nspts_func(long)
        ewpts = ewpts_func(short)
        if nspts == 1 and ewpts == 1:
            valid = single_centroid(geom)
        else:
            valid = point_distribution_func(geom, nspts, ewpts, Session=Session, **kwargs)
    elif ew == True:
        # Since this is an LS, we should place these diagonally from the 'lower left' to the 'upper right'
        nspts = ewpts_func(short)
        ewpts = nspts_func(long)
        if nspts == 1 and ewpts == 1:
            valid = single_centroid(geom)
        else:
            valid = point_distribution_func(geom, nspts, ewpts, Session=Session, **kwargs)
    else:
        log.warning('WTF Willy')
    return np.array(valid)


def alpha_shape(points, alpha):
    """
    Compute a convex hull from a set of points.

    credit: https://gist.github.com/dwyerk/10561690

    Parameters
    ----------

    points : np.array
             points in the format [[x0,y0], [x1, y,1], ... [xn, yn]]

    alpha : float
            Higher alphas creater a tighter the boundery around input points, alphas approaching 0 create a convex hull.
            Best to keep in the range (0, 1]


    Returns
    -------

    convex_hull : shapely.geometry.Polygon
                  Shapely polygon of the convex hull

    """
    tri = Delaunay(points)
    triangles = points[tri.vertices]
    a = ((triangles[:,0,0] - triangles[:,1,0]) ** 2 + (triangles[:,0,1] - triangles[:,1,1]) ** 2) ** 0.5
    b = ((triangles[:,1,0] - triangles[:,2,0]) ** 2 + (triangles[:,1,1] - triangles[:,2,1]) ** 2) ** 0.5
    c = ((triangles[:,2,0] - triangles[:,0,0]) ** 2 + (triangles[:,2,1] - triangles[:,0,1]) ** 2) ** 0.5
    s = ( a + b + c ) / 2.0
    areas = (s*(s-a)*(s-b)*(s-c)) ** 0.5
    circums = a * b * c / (4.0 * areas)

    # avoid devide by zero
    thresh = 1.0/alpha if alpha is not 0 else circums.max()
    filtered = triangles[circums < thresh]

    edge1 = filtered[:,(0,1)]
    edge2 = filtered[:,(1,2)]
    edge3 = filtered[:,(2,0)]
    edge_points = np.unique(np.concatenate((edge1,edge2,edge3)), axis = 0).tolist()
    m = geometry.MultiLineString(edge_points)
    triangles = list(cg.polygonize(m))
    return cascaded_union(triangles)


def rasterize_polygon(shape, vertices, dtype=bool):
    """
    Simple tool to convert poly into a boolean numpy array.

    source: https://stackoverflow.com/questions/37117878/generating-a-filled-polygon-inside-a-numpy-array

    Parameters
    ----------

    shape : tuple
            size of the array in (y,x) format

    vertices : np.array, list
               array of vertices in [[x0, y0], [x1, y1]...] format

    dtype : type
            datatype of output mask

    Returns
    -------

    mask : np.array
           mask with filled polygon set to true

    """
    def check(p1, p2, base_array):
        idxs = np.indices(base_array.shape) # Create 3D array of indices

        p1 = p1.astype(float)
        p2 = p2.astype(float)

        # Calculate max column idx for each row idx based on interpolated line between two points
        if p1[0] == p2[0]:
            max_col_idx = (idxs[0] - p1[0]) * idxs.shape[1]
            sign = np.sign(p2[1] - p1[1])
        else:
            max_col_idx = (idxs[0] - p1[0]) / (p2[0] - p1[0]) * (p2[1] - p1[1]) + p1[1]
            sign = np.sign(p2[0] - p1[0])

        return idxs[1] * sign <= max_col_idx * sign

    base_array = np.zeros(shape, dtype=dtype)  # Initialize your array of zeros

    fill = np.ones(base_array.shape) * True  # Initialize boolean array defining shape fill

    # Create check array for each edge segment, combine into fill array
    for k in range(vertices.shape[0]):
        fill = np.all([fill, check(vertices[k-1], vertices[k], base_array)], axis=0)

    log.info(fill.any())
    # Set all values inside polygon to one
    base_array[fill] = 1
    return base_array



import os

import numpy as np
import pandas as pd
from plio.io import io_hdf

from autocnet.utils import utils

def from_hdf(in_path, index=None, keypoints=True, descriptors=True):
    """
    For a given node, load the keypoints and descriptors from a hdf5 file. The
    keypoints and descriptors kwargs support returning only keypoints or descriptors.
    The index kwarg supports returning a subset of the data.

    Parameters
    ----------
    in_path : str
              handle to the file

    index : iterable
            an h5py accepted indexer to pull only a subset of the keypoints
            off disk. Default is None to pull all keypoints.

    keypoints : bool
                if True (default) return the keypoints

    descriptors : bool
                  if True (default) return the descriptors

    Returns
    -------
    keypoints : DataFrame
                A pandas dataframe of keypoints.

    descriptors : ndarray
                  A numpy array of descriptors
    """
    if isinstance(in_path, str):
        hdf = io_hdf.HDFDataset(in_path, mode='r')
    else:
        hdf = in_path

    outd = '/descriptors'
    outk = '/keypoints'

    if index is not None:
        index=np.asarray(index)

        # The indices into HDF have to be sorted lists. When indices get passed in
        # they are frequently ordered, so this pulls the data using the sorted
        # index and then reorders the data.
        i = np.argsort(index)
        ii = np.argsort(i)
        # Is is important to use sorted() so that an in-place sort is NOT used.
        if descriptors:
            desc = hdf[outd][index[i].tolist()]
            desc = desc[ii]
        if keypoints:
            raw_kps = hdf[outk][index[i].tolist()]
            raw_kps = raw_kps[ii]
    else:
        # Unlike numpy hdf does not handle NoneType as a proxy for `:`
        if descriptors:
            desc = hdf[outd][:]
        if keypoints:
            raw_kps = hdf[outk][:]
    
    if keypoints:
        index = raw_kps['index']
        clean_kps = utils.remove_field_name(raw_kps, 'index')
        columns = clean_kps.dtype.names

        allkps = pd.DataFrame(data=clean_kps, columns=columns, index=index)

    if isinstance(in_path, str):
        hdf = None

    if keypoints and descriptors:
        return allkps, desc
    elif keypoints:
        return allkps
    else:
        return desc


def to_hdf(out_path, keypoints=None, descriptors=None, key=None):
    """
    Save keypoints and descriptors to HDF at a given out_path at either
    the root or at some arbitrary path given by a key.

    Parameters
    ----------
    out_path : str
               to the HDF5 file

    keypoints : DataFrame
                Pandas dataframe of keypoints

    descriptors : ndarray
                  of feature descriptors

    key : str
          path within the HDF5 file.  If given, the keypoints and descriptors
          are save at <key>/keypoints and <key>/descriptors respectively.
    """
    # If the out_path is a string, access the HDF5 file
    if isinstance(out_path, str):
        hdf = io_hdf.HDFDataset(out_path, mode='a')
    else:
        hdf = out_path

    grps = list(hdf.keys())

    outd = '/descriptors'
    outk = '/keypoints'
    if descriptors is not None:
        # Strip the leading slash
        if outd[1:] in grps:
            del hdf[outd] # pragma: no cover

        hdf.create_dataset(outd,
                        data=descriptors,
                        compression=io_hdf.DEFAULT_COMPRESSION,
                        compression_opts=io_hdf.DEFAULT_COMPRESSION_VALUE)

    if keypoints is not None:
        if outk[1:] in grps:
            del hdf[outk]  # pragma: no cover

        hdf.create_dataset(outk,
                        data=hdf.df_to_sarray(keypoints.reset_index()),
                        compression=io_hdf.DEFAULT_COMPRESSION,
                        compression_opts=io_hdf.DEFAULT_COMPRESSION_VALUE)

    #except:
        #warnings.warn('Descriptors for the node {} are already stored'.format(self['image_name']))

    # If the out_path is a string, assume this method is being called as a singleton
    # and close the hdf file gracefully.  If an object, let the instantiator of the
    # object close the file
    if isinstance(out_path, str):
        del hdf

def from_npy(in_path):
    """
    Load keypoints and descriptors from a .npz file.

    Parameters
    ----------
    in_path : str
              PATH to the npz file

    Returns
    -------
    keypoints : DataFrame
                of keypoints

    descriptors : ndarray
                  of feature descriptors
    """
    nzf = np.load(in_path, allow_pickle=True)
    descriptors = nzf['descriptors']
    keypoints = pd.DataFrame(nzf['keypoints'], index=nzf['keypoints_idx'], columns=nzf['keypoints_columns'])

    return keypoints, descriptors

def to_npy(keypoints, descriptors, out_path):
    """
    Save keypoints and descriptors to a .npz file at some out_path

    Parameters
    ----------
    keypoints : DataFrame
                of keypoints

    descriptors : ndarray
                  of feature descriptors

    out_path : str
               PATH and filename to save the features
    """
    np.savez(out_path, descriptors=descriptors,
             keypoints=keypoints,
             keypoints_idx=keypoints.index,
             keypoints_columns=keypoints.columns)

def create_output_path(filename, outdir=None):
    """
    Given a filename for keypoints and descriptors, create an output
    directory with _kps.h5 appended.

    Parameters
    ----------
    filename : str
               The filename or full path
    outdir : str
             An optional output path

    Returns
    -------
    outh5 : str
            Path of the output h5 file
    """
    image_name = os.path.basename(filename)                                    
    image_path = os.path.dirname(filename)                                     
                                                                                   
    if outdir is None:                                                             
        outh5 = os.path.join(image_path, image_name + '_kps.h5')                   
    else:                                                                          
        outh5 = os.path.join(outdir, image_name + '_kps.h5')                       
                                                                                   
    return outh5 
